import numpy as np
import cv2 as cv
from skimage.util import img_as_ubyte, img_as_uint
from skimage import io
import glob
import os.path


def load_image(path_in, img_bitdepth='uint8', quiet_in=False):
    """ 
    Loads an image file using OpenCV routines. Color images will be
    converted to grayscale, and uint16 images will linearly scaled to
    uint8. Only uint16 and uint8 image data types are supported. 
    
    ---- INPUT ARGUMENTS ----
    path_in: String that contains the file path (and name and extension)
        of the image being loaded.

    img_bitdepth [string]: Bit depth for the reader to use. Either
        of unsigned 8-bit (uint8) or unsigned 16-bit (uint16) are
        nominally supported. 8-bit is the default (and what the rest of
        the codes currently expect).
    
    quiet_in: An optional boolean that is by default set to False. Set 
        to True to prevent outputting any messages.

    ---- RETURNED ---- 
    [img, img_prop]: List of length two. If the image failed to load 
        correctly, this list will contain just None objects. In more 
        detail, these returned items are described below.

        img: OpenCV's Mat class that describes the image data via a
            dense n-dimensional array of numbers. Conventional numpy 
            operations can be safely applied to img.

        img_prop: List of length three that describes the image
            properties. More details given below.

            img_prop[0]: Total number of pixels in the image.
            img_prop[1]: A tuple of length two that provides the number 
                of rows and columns of pixels. 
            img_prop[2]: A string that confirms the image data type. 
                Since load_image(...) converts all images to 'uint8', 
                this will  always be equal to 'uint8'.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings may be
    printed to standard output.
    """

    # ---- Start Local Copies ----
    # Forcing Python to create a new string variable in memory
    # File path to the image file
    file_path = (path_in + '.')[:-1]

    # Makes forward slashes into backward slashes for W10, and also makes the
    # directory string all lowercase
    file_path = os.path.normcase(file_path)
    
    quiet = quiet_in
    # ---- End Local Copies ----

    # Load a image in possibly uint16 or uint8 format (in grayscale).
    #img = cv.imread(file_path, cv.IMREAD_ANYDEPTH)
    # Note, OpenCV opens color images in a BGR convention, not RGB. To swap
    # the channels, which is necessary to plot color images in matlplotlib,
    # you can use the following function: cv.cvtColor(image,
    # cv.COLOR_RGB2BGR). However, this script converts all color images into
    # grayscale, so this does not actually apply here.

    # OpenCV could not handle image rotations correctly for 16-bit images.
    # Using SciKit-Image instead.
    img = io.imread(file_path, as_gray=True)

    # Catch the case that the file path was incorrect, or the image is
    # corrupted
    if img is None:
        return [None, None]
    
    # Total number of pixels
    image_size = img.size
    
    # If the image is color, img.shape returns a tuple containing the number
    # of  rows, columns, and channels. If the image is grayscale, this returns
    # a tuple containing the number of rows and columns (no color channels)
    image_shape = img.shape
    
    # Returns the image data type (e.g., uint8)
    image_data_type = img.dtype

    # If uint16, convert to uint8 image data type
    if img_bitdepth == "uint8":
        # Use SciKit-Image to be more robust; is should handle n-d arrays
        img = img_as_ubyte(img)
        
    elif img_bitdepth == "uint16":
        # Use SciKit-Image to be more robust
        img = img_as_uint(img)
        
    else:
        if not quiet:
            print("\nWARNING: Unsupported bit depth detected. Defaulting to 8-bit")
        img = img_as_ubyte(img)
    
    # Update image parameters in case any changes were made
    image_size = img.size
    image_shape = img.shape
    image_data_type = img.dtype
    img_prop = [image_size, image_shape, image_data_type]

    if not quiet:
        print(f"\nSuccessfully imported image: {file_path}")

    # Return the image objects and image properties
    return [img, img_prop]


def load_multipage_image(path_in, indices_in=[], bigtiff=False,\
                         img_bitdepth_in="uint8", flipz=False,
                         quiet_in=False):
    """
    Read a multipage image (e.g. monolithic tiff stack) from file using
    the skimage library.
    
    ---- INPUT ARGUMENTS ----
    path_in [string]: Path to tiff file to load
    
    indices_in: An optional tuple of either length 1 or length 2. If 
        length 1, then it should contain a positive integer 
        corresponding to the number of images to be kept centered about
        the middle image of the image sequence. If length 2, then the
        first and second elements will be used directly to slice the
        list corresponding to the image sequence. For example, (0, 100)
        would keep the first 100 images. Fair warning, if you provide
        invalid indices, you will get an error.

    bigtiff [bool]: If False, the OpenCV multi-page TIFF function will
        be used to import the image. This is fine for standard TIFF 
        files. However, if the TIFF file is saved with a less 
        conventional format/header, such as for BigTIFF or ImageJ 
        Hyperstack formats, then this should be set to True in order to
        use a different importer. Note, these alternative TIFF formats
        should be used anytime the TIFF file is larger than 4 GB. When
        set to True, the tifffile library will be used via a plugin 
        within the Sci-Kit Image library.
    
    img_bitdepth_in [string]: Bit depth for the reader to use. Either
        of unsigned 8-bit (uint8) or unsigned 16-bit (uint16) are nominally
        supported. 8-bit is the default (and what the rest of the codes
        currently expect).

    flipz [bool]: When True, the images are reversed in the image stack,
        which can be thought of as flipping the image stack along the Z-
        direction. This occurs after any images have been removed from
        'indices_in' above. When False, the original order of the image
        stack is maintained. If the first image corresponds to the 
        bottom of a part, then in general, flipz should be True so as
        to ensure a right-handed coordinate system. Positive X will be
        along ascending column indices, positive Y will be along 
        ascending row indices, and positive Z will be along DESCENDING
        image indices.
        
    quiet_in [bool]: Set to true to suppress any output dialog
    
    --- RETURNED ---
    [img, img_prop]

    img: OpenCV Mats n-d array containing images (can be operated on
        like a numpy array). Its shape is [num_images, num_rows, 
        num_cols].

    img_prop: List of length three that describes the image
            properties.
        img_prop[0]: Total number of pixels in the image.
        img_prop[1]: A tuple of length three that provides the number 
            of rows, columns, and pages of pixels. 
        img_prop[2]: A string that confirms the image data type. 
            Since load_image(...) converts all images to 'uint8', 
            this will  always be equal to 'uint8'.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings may be
    printed to standard output.
    
    Warning: this loads the full image stack and then slices out the
        desired parts defined in indices_in. There maybe be a way to
        do this better ...
    
    """
    
    # ---- Start Local Copies ----
    # Forcing Python to create a new string variable in memory
    # File path to the image file
    file_path = (path_in + '.')[:-1]

    # Makes forward slashes into backward slashes for W10, and also makes the
    # directory string all lowercase
    file_path = os.path.normcase(file_path)
    
    quiet = quiet_in

    img_bitdepth = img_bitdepth_in.lower()

    indices_keep = list(indices_in) # Ex: (0, 100) keeps first 100 images

    # ---- End Local Copies ----


    if not quiet:
        print(f"\nImporting multi-page TIFF stack:\n   {file_path}")

    imgs = io.imread(fname=file_path, plugin='tifffile')

    # Catch the case that the file path was incorrect, or the image is
    # corrupted
    if (len(imgs) == 0) and (not quiet):
        print(f"\nERROR: Image not loaded from filepath:\n  {file_path}\n")
        # Use "raise Exception()" here too?
        return [None, None]

    imgs = np.array(imgs) # Convert to Numpy array if not already
    
    # Contains the number of row, columns, and pages in a 3D image
    num_imgs = imgs.shape[0]
    num_rows = imgs.shape[1]
    num_cols = imgs.shape[2]
    image_shape = [num_imgs, num_rows, num_cols]

    # Returns the image data type (e.g., uint8)
    image_data_type = imgs[0].dtype
    
    # If uint16, convert to uint8 image data type
    if img_bitdepth == "uint8":
        # Use SciKit-Image to be more robust; is should handle n-d arrays
        imgs = img_as_ubyte(imgs)
        
    elif img_bitdepth == "uint16":
        # Use SciKit-Image to be more robust
        imgs = img_as_uint(imgs)
        
    else:
        if not quiet:
            print("\nWARNING: Unsupported bit depth detected. Defaulting to 8-bit")
        imgs = img_as_ubyte(imgs)
    
    if len(indices_in) == 0:
        # return full image
        indx_bounds = [0, num_imgs]
        if not quiet:
            print("\nReturning full image")

    elif len(indices_in) == 1:
        n_keep = indices_keep[0]
        cur_img_len = num_imgs # Total number of images

        if n_keep == 0: # Keeping zero images doesn't make sense
            n_keep = 1  # Will just keep the middle image then
        elif n_keep > cur_img_len: # Also doesn't make sense
            n_keep = cur_img_len   # So, keep them all

        # Index of middle element, rounds down if even
        if cur_img_len % 2 == 0: # If even
            mid_indx = int(cur_img_len/2) - 1
        else: # If odd
            mid_indx = int(cur_img_len/2)

        # Force the desired number of images to keep to be odd for now
        keep_even = False
        if n_keep % 2 == 0:
            keep_even = True
            n_keep -= 1

        # Number of images to keep above and below the middle index
        indx_rad = int((n_keep - 1)/2)

        # Index bounds for the images to be kept
        indx_bounds = [mid_indx - indx_rad, mid_indx + indx_rad + 1]

        # If originally even number of images to keep, correct it here
        if keep_even:
            indx_bounds[1] = indx_bounds[1] + 1

    # If there's two numbers, user gave the upper and lower numbers
    elif len(indices_in) == 2:
        indx_bounds = [indices_keep[0], indices_keep[1]]

    else:
        indx_bounds = [0, num_imgs]
        if not quiet: 
            print("\nWARNING: 'indices_in' is malformed. Returning full image")
    
    imgs = imgs[indx_bounds[0]:indx_bounds[1],:,:]

    # Reverse the order of the image stack
    if flipz:
        if not quiet:
            print(f"\nReversing the order of the image stack (i.e.," \
                + " flipping the Z-direction)...")
        imgs = np.flip(imgs, axis=0)

    # Update image parameters in case any changes were made
    image_size = imgs.size

    num_imgs = imgs.shape[0]
    num_rows = imgs.shape[1]
    num_cols = imgs.shape[2]
    image_shape = [num_imgs, num_rows, num_cols]

    image_data_type = imgs.dtype
    img_prop = [image_size, image_shape, image_data_type]

    if not quiet:
        print(f"\nSuccessfully imported image: {file_path}")

    # Return the image objects and image properties
    return [imgs, img_prop]


def read_pgm(filename_in, transpose=False, ASCII=True):
    """
    Read ASCII PGM file with magic number P2 using Ed's conventions.
    First line should contain "P2". Second line contains two integers:
    in the standard format, the first number should be the number of
    columns and second number should be the number of rows. Finally, the
    flattened array of pixel values are given -- one per line. Ed has
    them flattened in row-major order (C convention). The input, 
    "filename_in", is a string path to the ASCII image file that will be
     imported. The returned image is a Numpy 2D matrix with data type
     UINT8 (i.e., grayscale 0 - 255). Inputs and outputs are similar in
     style to the above function, load_image(...). Note, to ready 
    """

    # Force a local copy of this string
    file_name = (filename_in + '.')[:-1]
    file_name = os.path.normcase(file_name)

    if ASCII:
        with open(file_name) as f:
            lines = f.readlines()

        # Ignores commented lines
        for l in list(lines):
            if l[0] == '#':
                lines.remove(l)

        # Makes sure it is ASCII format (P2)
        assert lines[0].strip() == 'P2' 

        # Converts data to a list of integers
        data = []
        for line in lines[1:]:
            data.extend([int(c) for c in line.split()])
        
        if transpose:
            img1_shape = (data[0], data[1]) # (num of rows, num of cols)

        else:
            # Standard PGM file format
            img1_shape = (data[1], data[0]) # (num of rows, num of cols)

        img1_max = data[2]
        img1 = np.array(data[3:])
        img1 = np.reshape(img1, img1_shape, order='C')

        # Convert the data to a grayscale image from 0 - 255
        if np.amin(img1) < 0:
            img1 = img1 - np.amin(img1)

        if (img1_max > 255) or (img1_max != 255):
            ratio = img1_max/255.0
            img1 = (img1/ratio).astype(np.uint8)

        img1 = img_as_ubyte(img1)

    else: 
        # Import as binary format. This must be in the standard format.
        
        img1 = io.imread(file_name)
        img1 = img_as_ubyte(img1)

        img1_shape = img1.shape
    
    return (img1, img1_shape)

def load_image_seq(path_in, file_name_in='', img_bitdepth_in='uint8', 
    indices_in=(), flipz=False):
    """
    Loads an image sequence into memory in a single batch operation.
    This is done by repeatedly using load_image(...). Hence, the list
    of returned images will be grayscale and of type uint8. It is 
    assumed all of the images are in the same directory. Images are 
    sorted in ascending order (alphabetical).

    ---- INPUT ARGUMENTS ---- 
    path_in: String that contains the directory path.

    file_name_in: An optional string that is taken to be a substring of
        the image files to be imported. For  example, assume the
        following images are in a folder: img001.tif,  img002.tif,
        img003.tif, and other_img.png. Then, file_name_in can be set
        equal to "img" to ensure that only the .tif files are imported.
        By default, it is an empty string, in which case, every image
        in the directory will be imported.

    img_bitdepth_in [string]: Bit depth for the reader to use. Either
        of unsigned 8-bit (uint8) or unsigned 16-bit (uint16) are
        nominally supported. 8-bit is the default (and what the rest of
        the codes currently expect).

    indices_in: An optional tuple of either length 1 or length 2. If 
        length 1, then it should contain a positive integer 
        corresponding to the number of images to be kept centered about
        the middle image of the image sequence. If length 2, then the
        first and second elements will be used directly to slice the
        list corresponding to the image sequence. For example, (0, 100)
        would keep the first 100 images. Fair warning, if you provide
        invalid indices, you will get an error.

    flipz [bool]: When True, the images are reversed in the image stack,
        which can be thought of as flipping the image stack along the Z-
        direction. This occurs after any images have been removed from
        'indices_in' above. When False, the original order of the image
        stack is maintained. If the first image corresponds to the 
        bottom of a part, then in general, flipz should be True so as
        to ensure a right-handed coordinate system. Positive X will be
        along ascending column indices, positive Y will be along 
        ascending row indices, and positive Z will be along DESCENDING
        image indices.

    ---- RETURNED ---- 
    [[imgs], [img_names]]: List of length two. If the images failed to
        load correctly, this list will contain just None objects. See 
        below for more details.

        [imgs]: A 3D Numpy array that contains OpenCV's Mat class  
            objects, which are 2D Numpy arrays in of themselves. 
            Conventional Numpy operations can be safely applied to each
            entry of imgs. Images are converted to grayscale and uint8
            automatically. The shape of imgs will be (num_images,
            num_rows, num_cols).

        [img_names]: A list of strings that describe the file paths used
            to import every image. 

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings may be
    printed to standard output.
    """

    # ---- Start Local Copies ----
    # Forcing Python to create a new string variable in memory
    # Path to the directory containing images
    dir_path = (path_in + '.')[:-1]
    if not dir_path.endswith( ('\\', '/') ):
        dir_path = dir_path + '/'

    # Makes forward slashes into backward slashes for W10, and also makes the
    # directory string all lowercase
    dir_path = os.path.normcase(dir_path) 
    
    # Substring of the filenames to be imported
    file_name = (file_name_in + '.')[:-1] 
    file_name = os.path.normcase(file_name)

    # If length 1, then total number of images to keep about the center
    # If length 2, then lower and upper bounds of the images to be kept
    indices_keep = list(indices_in) # Ex: (0, 100) keeps first 100 images
    # ---- End Start Local Copies ----

    # Reduced list (actually tuple) of supported image file extensions
    file_ext = (".tif", ".tiff", ".png", ".jpg", ".jpeg", ".jp2", ".bmp",
                ".dib", ".pbm", ".ppm", ".pgm", ".pnm")

    # Look for files that contain file_name in their names
    img_names = [] # A list of strings for the image file paths
    if file_name:
        files = glob.glob(dir_path + '*') # All files in directory
        if files:
            for cur_name in files:
                # Makes forward slashes into backward slashes for W10, and
                # also makes the directory string all lowercase
                cur_name = os.path.normcase(cur_name)
                if file_name in cur_name:
                    # Also ensure this is actually an image file
                    if (cur_name.lower()).endswith(file_ext):
                        img_names.append(cur_name)

    # Otherwise, import all common image file types (case insensitive)  
    else:
        files = glob.glob(dir_path + '*') # All files in directory
        if files:
            for cur_name in files:
                # Makes forward slashes into backward slashes for W10, and
                # also makes the directory string all lowercase
                cur_name = os.path.normcase(cur_name)
                # Only keep image files
                if (cur_name.lower()).endswith(file_ext):
                    img_names.append(cur_name)

    # Ensure that the list of names is not empty
    if not img_names:
        print(f"\nNo image files were found in {dir_path}")
        return [None, None]

    # The Glob module does not guarantee that the results will be sorted.
    img_names = list(dict.fromkeys(img_names)) # Removes duplicates
    img_names.sort() # Asecending order based on the string file names

    if indices_keep: # If not empty
        # If not already, force the data to be positive and of type integer
        for count, element in enumerate(indices_keep):
            indices_keep[count] = int(abs(element))

        if len(indices_keep) == 1: # Select a number of images from the center
            n_keep = indices_keep[0]
            cur_img_len = len(img_names) # Total number of images

            if n_keep == 0: # Keeping zero images doesn't make sense
                n_keep = 1  # Will just keep the middle image then
            elif n_keep > cur_img_len: # Also doesn't make sense
                n_keep = cur_img_len   # So, keep them all

            # Index of middle element, rounds down if even
            if cur_img_len % 2 == 0: # If even
                mid_indx = int(cur_img_len/2) - 1
            else: # If odd
                mid_indx = int(cur_img_len/2)

            # Force the desired number of images to keep to be odd for now
            keep_even = False
            if n_keep % 2 == 0:
                keep_even = True
                n_keep -= 1

            # Number of images to keep above and below the middle index
            indx_rad = int((n_keep - 1)/2)

            # Index bounds for the images to be kept
            indx_bounds = [mid_indx - indx_rad, mid_indx + indx_rad + 1]

            # If originally even number of images to keep, correct it here
            if keep_even:
                indx_bounds[1] = indx_bounds[1] + 1

            # Perform the slice operation
            img_names = img_names[indx_bounds[0]:indx_bounds[1]]

        elif len(indices_keep) == 2: # Expecting lower and upper indices
            ind_low = indices_keep[0]
            ind_upp = indices_keep[1]

            img_names = img_names[ind_low:ind_upp]

        else:
            print(f"\nInvalid indices to slice the array of images. Provided"\
                f" input was: {indices_keep}\nExpecting a tuple of length "\
                f"1 or 2. For example:\n    1) (200,) --> Keep 200 images "\
                f"about the center.\n    2) (0, 100) --> Keep the first 100 "\
                f"images")

    # Number of images in the finalized list of file paths
    num_imgs = len(img_names)

    # [img_props]: A list containing the image properties for each 
    # imported image. Each entry in img_props is a list in of 
    # itself. More details given below.
    # 
    # img_prop[m][0]: Total number of pixels in the m-th image.
    # img_prop[m][1]: A tuple of length two that provides the  
    #     number of rows and columns of pixels for the m-th image. 
    # img_prop[m][2]: A string that confirms the image data type 
    #     for m-th image. Since load_image(...) converts all 
    #     images to 'uint8', this will always be equal to 'uint8'.
    img_props = []
    imgs = [] # Will be converted to Numpy array later
    counter = 1
    print(f"\nBeginning to import {num_imgs} images...")
    for cur_name in img_names:
        # Now actually call OpenCV routines to import the images
        # Use the defined function from above to import images in gray scale
        # and as uint8. Also, use quiet-mode (no outputs to the terminal)
        cur_img, cur_img_prop = load_image(cur_name, 
            img_bitdepth=img_bitdepth_in, quiet_in=True)

        if cur_img is None:
            print(f"\nFailed to import image: {cur_name}")
            return [None, None]

        # Store the image objects and their properties in lists
        imgs.append(cur_img)
        img_props.append(cur_img_prop)

        # Write out updates for the user
        if (counter%20) == 0:
            print(f"    Currently imported {counter}/{num_imgs}...")
        
        counter += 1

    print(f"\nSuccessfully imported {num_imgs} images!")

    # Check to make sure all the images are the same shape
    first_img_shape = img_props[0][1] # Tuple of the image shape (rows, cols)
    cur_index = 0
    for cur_img_prop in img_props:
        cur_shape = cur_img_prop[1]

        if cur_shape != first_img_shape:
            print(f"\nWARNING! Not all imported images are of the same shape")
            print(f"\nFirst image shape: {first_img_shape}")
            print(f"\nImage shape of {img_names[cur_index]}: {cur_shape}")

        cur_index += 1

    # Convert from a Python list of Numpy arrays to a fully 3D Numpy array
    imgs = np.array(imgs)

    # Reverse the order of the image stack
    if flipz:
        print(f"\nReversing the order of the image stack (i.e.," \
            + " flipping the Z-direction)...")
        imgs = np.flip(imgs, axis=0)

    # Return the image objects, image properties, and string file names
    return [imgs, img_names]


def load_image_seq_ASCII(path_in, file_name_in='', indices_in=(), 
    transpose=False, ASCII=True):
    """
    Loads an image sequence into memory in a single batch operation.
    This is done by repeatedly using load_image(...). Hence, the list
    of returned images will be grayscale and of type uint8. It is 
    assumed all of the images are in the same directory. Images are 
    sorted in ascending order (alphabetical). This function is 
    specifically for Ed's ASCII PGM image file type.

    ---- INPUT ARGUMENTS ---- 
    path_in: String that contains the directory path.

    file_name_in: An optional string that is taken to be a substring of
        the image files to be imported. For  example, assume the
        following images are in a folder: img001.pgm,  img002.pgm,
        img003.pgm, and other_img.png. Then, file_name_in can be set
        equal to "img" to ensure that only the .pgm files are imported.
        By default, it is an empty string, in which case, every image
        in the directory will be imported.

    indices_in: An optional tuple of either length 1 or length 2. If 
        length 1, then it should contain a positive integer 
        corresponding to the number of images to be kept centered about
        the middle image of the image sequence. If length 2, then the
        first and second elements will be used directly to slice the
        list corresponding to the image sequence. For example, (0, 100)
        would keep the first 100 images. Fair warning, if you provide
        invalid indices, you will get an error.

    transpose: A boolean that determines how to interpret the shape of 
        each image. A standard PGM file format will include a line with
        two integers corresponding to the shape of the image being
        imported. If transpose=False, then the first integer is taken to
        be the number of columns and the second integer the number of 
        rows; this is the standard format. If transpose=True, then the
        first integer is taken to be the number of rows and the second
        integer the number of columns.

    ASCII: A boolean to determine if the images should be imported as
        ASCII files or as binary images. Set to True for ASCII files,
        and set to False for binary files.

    ---- RETURNED ---- 
    [[imgs], [img_names]]: List of length two. If the images failed to
        load correctly, this list will contain just None objects. See 
        below for more details.

        [imgs]: A 3D Numpy array that contains OpenCV's Mat class  
            objects, which are 2D Numpy arrays in of themselves. 
            Conventional Numpy operations can be safely applied to each
            entry of imgs. Images are converted to grayscale and uint8
            automatically. The shape of imgs will be (num_images,
            num_rows, num_cols).

        [img_names]: A list of strings that describe the file paths used
            to import every image. 

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings may be
    printed to standard output.
    """

    # ---- Start Local Copies ----
    # Forcing Python to create a new string variable in memory
    # Path to the directory containing images
    dir_path = (path_in + '.')[:-1]
    if not dir_path.endswith( ('\\', '/') ):
        dir_path = dir_path + '/'

    # Makes forward slashes into backward slashes for W10, and also makes the
    # directory string all lowercase
    dir_path = os.path.normcase(dir_path)

    # Substring of the filenames to be imported
    file_name = (file_name_in + '.')[:-1] 
    file_name = os.path.normcase(file_name)

    # If length 1, then total number of images to keep about the center
    # If length 2, then lower and upper bounds of the images to be kept
    indices_keep = list(indices_in) # Ex: (0, 100) keeps first 100 images
    # ---- End Start Local Copies ----

    # Reduced list (actually tuple) of supported image file extensions
    file_ext = (".tif", ".tiff", ".png", ".jpg", ".jpeg", ".jp2", ".bmp",
                ".dib", ".pbm", ".ppm", ".pgm", ".pnm")

    # Look for files that contain file_name in their names
    img_names = [] # A list of strings for the image file paths
    if file_name:
        files = glob.glob(dir_path + '*') # All files in directory
        if files:
            for cur_name in files:
                # Makes forward slashes into backward slashes for W10, and
                # also makes the directory string all lowercase
                cur_name = os.path.normcase(cur_name)
                if file_name in cur_name:
                    # Also ensure this is actually an image file
                    if (cur_name.lower()).endswith(file_ext):
                        img_names.append(cur_name)

    # Otherwise, import all common image file types (case insensitive)  
    else:
        files = glob.glob(dir_path + '*') # All files in directory
        if files:
            for cur_name in files:
                # Makes forward slashes into backward slashes for W10, and
                # also makes the directory string all lowercase
                cur_name = os.path.normcase(cur_name)
                # Only keep image files
                if (cur_name.lower()).endswith(file_ext):
                    img_names.append(cur_name)

    # Ensure that the list of names is not empty
    if not img_names:
        print(f"\nNo image files were found in {dir_path}")
        return [None, None]

    # The Glob module does not guarantee that the results will be sorted.
    img_names = list(dict.fromkeys(img_names)) # Removes duplicates
    img_names.sort() # Asecending order based on the string file names

    if indices_keep: # If not empty
        # If not already, force the data to be positive and of type integer
        for count, element in enumerate(indices_keep):
            indices_keep[count] = int(abs(element))

        if len(indices_keep) == 1: # Select a number of images from the center
            n_keep = indices_keep[0]
            cur_img_len = len(img_names) # Total number of images

            if n_keep == 0: # Keeping zero images doesn't make sense
                n_keep = 1  # Will just keep the middle image then
            elif n_keep > cur_img_len: # Also doesn't make sense
                n_keep = cur_img_len   # So, keep them all

            # Index of middle element, rounds down if even
            if cur_img_len % 2 == 0: # If even
                mid_indx = int(cur_img_len/2) - 1
            else: # If odd
                mid_indx = int(cur_img_len/2)

            # Force the desired number of images to keep to be odd for now
            keep_even = False
            if n_keep % 2 == 0:
                keep_even = True
                n_keep -= 1

            # Number of images to keep above and below the middle index
            indx_rad = int((n_keep - 1)/2)

            # Index bounds for the images to be kept
            indx_bounds = [mid_indx - indx_rad, mid_indx + indx_rad + 1]

            # If originally even number of images to keep, correct it here
            if keep_even:
                indx_bounds[1] = indx_bounds[1] + 1

            # Perform the slice operation
            img_names = img_names[indx_bounds[0]:indx_bounds[1]]

        elif len(indices_keep) == 2: # Expecting lower and upper indices
            ind_low = indices_keep[0]
            ind_upp = indices_keep[1]

            img_names = img_names[ind_low:ind_upp]

        else:
            print(f"\nInvalid indices to slice the array of images. Provided"\
                f" input was: {indices_keep}\nExpecting a tuple of length "\
                f"1 or 2. For example:\n    1) (200,) --> Keep 200 images "\
                f"about the center.\n    2) (0, 100) --> Keep the first 100 "\
                f"images")

    # Number of images in the finalized list of file paths
    num_imgs = len(img_names)

    # [img_props]: A list containing the image properties for each 
    # imported image. Each entry in img_props is a list in of 
    # itself. More details given below.
    # 
    # img_prop[m][0]: Total number of pixels in the m-th image.
    # img_prop[m][1]: A tuple of length two that provides the  
    #     number of rows and columns of pixels for the m-th image. 
    # img_prop[m][2]: A string that confirms the image data type 
    #     for m-th image. Since load_image(...) converts all 
    #     images to 'uint8', this will always be equal to 'uint8'.
    img_props = []
    imgs = [] # Will be converted to Numpy array later
    counter = 1
    print(f"\nBeginning to import {num_imgs} images...")
    for cur_name in img_names:
        # Now actually call OpenCV routines to import the images
        # Use the defined function from above to import images in gray scale
        # and as uint8. Also, use quiet-mode (no outputs to the terminal)

        [cur_img, temp1] = read_pgm(cur_name, transpose=transpose, 
            ASCII=ASCII)
        cur_img_prop = [cur_img.size, cur_img.shape, cur_img.dtype]

        #cur_img, cur_img_prop = load_image(cur_name, quiet_in=True)
        #if cur_img is None:
        #    print(f"\nFailed to import image: {cur_name}")
        #    return [None, None]

        # Store the image objects and their properties in lists
        imgs.append(cur_img)
        img_props.append(cur_img_prop)

        # Write out updates for the user
        if (counter%20) == 0:
            print(f"    Currently imported {counter}/{num_imgs}...")
        
        counter += 1

    print(f"\nSuccessfully imported {num_imgs} images!")

    # Check to make sure all the images are the same shape
    first_img_shape = img_props[0][1] # Tuple of the image shape (rows, cols)
    cur_index = 0
    for cur_img_prop in img_props:
        cur_shape = cur_img_prop[1]

        if cur_shape != first_img_shape:
            print(f"\nWARNING! Not all imported images are of the same shape")
            print(f"\nFirst image shape: {first_img_shape}")
            print(f"\nImage shape of {img_names[cur_index]}: {cur_shape}")

        cur_index += 1

    # Convert from a Python list of Numpy arrays to a fully 3D Numpy array
    imgs = np.array(imgs)

    # Return the image objects, image properties, and string file names
    return [imgs, img_names]


def save_image(img_in, path_in, compression=False, quiet_in=False):
    """ 
    Saves an OpenCV image to the hard drive.
    
    ---- INPUT ARGUMENTS ----
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 

    path_in: String that contains the file path (and name and extension)
        of the image being saved.

    compression: A boolean that determines if the tiff images should
        be compressed. If True, LZW compression is done. If False,
        no compression is performed. This compression flag is only
        used for saving tif (or tiff) images.

    quiet_in: An optional boolean that is by default set to False. Set 
        to True to prevent outputting any messages.

    ---- RETURNED ---- 
    retval: A boolean value. Returns True if img_in was successfully
        saved using the file path given by path_in. If img_in failed
        to save, retval will be False.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. An image file is written
    to the hard drive. Strings may be printed to standard output.
    """

    # Make local copies
    img = img_in # No need for a deep copy. This image won't be altered

    # Forcing Python to create a new string variable in memory
    # Path to the directory to save the images
    file_path = (path_in + '.')[:-1]

    # Makes forward slashes into backward slashes for W10, and also makes the
    # directory string all lowercase
    file_path = os.path.normcase(file_path)
    compress_bool = compression
    quiet = quiet_in

    # Reduced list (actually tuple) of supported image file extensions
    file_ext = (".tif", ".tiff", ".png", ".jpg", ".jpeg", ".jp2", ".bmp",
                ".dib", ".pbm", ".ppm", ".pgm", ".pnm")

    retval = False

    try:
        if compress_bool:
            # LZW compression if True
            retval = cv.imwrite(file_path, img, 
                params=(cv.IMWRITE_TIFF_COMPRESSION, 5))

        else:
            # No compression if False
            retval = cv.imwrite(file_path, img, 
                params=(cv.IMWRITE_TIFF_COMPRESSION, 1))

    except:
        if not quiet:
            print(f"\nERROR: Encountered an error trying to save {file_path}")

        if not (file_path.lower()).endswith(file_ext):

            if not quiet:
                print(f"Detected invalid file extension. Attempting to "\
                "save as a '.tif' file...")

            file_path = file_path + ".tif"

            try:
                if compress_bool:
                    # LZW compression if True
                    retval = cv.imwrite(file_path, img, 
                        params=(cv.IMWRITE_TIFF_COMPRESSION, 5))

                else:
                    # No compression if False
                    retval = cv.imwrite(file_path, img, 
                        params=(cv.IMWRITE_TIFF_COMPRESSION, 1))

            except:
                if not quiet:
                    print(f"FAILED to save {file_path}")

    if retval:
        if not quiet_in:
            print(f"\nSuccessfully saved {file_path}")

    return retval


def save_image_seq(imgs_in, dir_in_path, file_name_in, 
    index_start_in=0, compression=False):
    """ 
    Saves a list of OpenCV images to the hard drive.
    
    ---- INPUT ARGUMENTS ----
    [imgs_in]: A 3D Numpy array containing images, each represented by
        a Numpy array, which corresponds to OpenCV's single-channel 
        uint8 (grayscale) image data type. The shape of imgs_in should
        be (num_images, num_rows, num_cols).

    dir_in_path: String that contains the directory path of the image
        being saved. The directory path will be normalized and made 
        lowercase, so forward or backward slashes are acceptable.

    file_name_in: String containing the name to be used for saving all
        of the processed images. The image names will automatically be
        appended with a 4-digit number to denote the correct sequence.
        Be sure to include the desired filetype, else by default, the
        images will be saved as '.tif' files. For example, if you had
        three images to save and chose name_out_substr = "out.tif", 
        then the saved images would be 'out_0000.tif', 'out_0001.tif',
        and 'out_0002.tif'.

    index_start_in: An integer that is used as initial value for the
        appended number to the saved image name. For example, if you had
        three images to save and chose name_out_substr = "out.tif" and
        index_start_in = 10,  then the saved images would be
        'out_0010.tif', 'out_0011.tif', and 'out_0012.tif'.

    compression: A boolean that determines if the tiff images should
        be compressed. If True, LZW compression is done. If False,
        no compression is performed. This compression flag is only
        used for saving tif (or tiff) images.

    ---- RETURNED ---- 
    Nothing is returned    

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Image files are written
    to the hard drive. Strings may be printed to standard output.
    """
    
    # Make local copies
    imgs = imgs_in # No need for a deep copy. This image won't be altered

    # Forcing Python to create a new string variable in memory
    file_path = (dir_in_path + '.')[:-1]
    if not file_path.endswith( ('\\', '/') ):
        file_path = file_path + '/'

    file_name = (file_name_in + '.')[:-1]
    file_name = os.path.normcase(file_name)

    # Makes forward slashes into backward slashes for W10, and also makes the
    # directory string all lowercase
    file_path = os.path.normcase(file_path)
    index_start = int(index_start_in)
    compress_bool = compression

    # Reduced list (actually tuple) of supported image file extensions
    file_ext3 = (".tif", ".png", ".jpg", ".jp2", ".bmp", ".dib", ".pbm",
                 ".ppm", ".pgm", ".pnm")
    file_ext4 = (".tiff", ".jpeg")

    # Check for the correct file extension and save it
    if (file_name.lower()).endswith(file_ext3):
        file_ext = file_name[-4:] # Does inclue '.'
        name_substr = file_name[:-4] # Does not inluce '.'

    elif (file_name.lower()).endswith(file_ext4):
        file_ext = file_name[-5:] # Does inclue '.'
        name_substr = file_name[:-5] # Does not inluce '.'

    else:
        print(f"\nWarning: No supported file extension found in "\
            f"'{file_name}'\n    Defaulting to '.tif' file extension.")
        name_substr = file_name
        file_ext = ".tif"

    num_imgs = imgs.shape[0]
    img_ii = index_start
    print(f"\nBeginning to save {num_imgs} images to:\n    {file_path}")
    for img_count, cur_img in enumerate(imgs):

        if num_imgs < 1E4:
            img_str_indx = "_" + str(img_ii).zfill(4)

        elif 1E4 <= num_imgs < 1E5:
            img_str_indx = "_" + str(img_ii).zfill(5)

        elif 1E5 <= num_imgs < 1E6:
            img_str_indx = "_" + str(img_ii).zfill(6)

        elif 1E6 <= num_imgs < 1E7:
            img_str_indx = "_" + str(img_ii).zfill(7)

        else:
            print("\nERROR: Over 1E7 images not currently supported in "\
                "save_image_seq(...). \nPlease use smaller batches.")
            return

        # Construct the full file path for the current image
        # (directory_path) + (file_name_substring) + (file_index) + 
        # (file_extension)
        cur_file_name = file_path + name_substr + img_str_indx + file_ext

        save_flag = save_image(cur_img, cur_file_name, 
            compression=compress_bool, quiet_in=True)

        if not save_flag:
            # I wonder if using a "raise Exception()" would be better
            # than a simple print statement here? That way one could call
            # it more robustly in a try...except block, right?
            print(f"\nERROR: Failed to save {cur_file_name}")

        elif ((img_count+1)%20) == 0:
            print(f"    Currently saved {img_count+1}/{num_imgs}...")

        img_ii += 1

    if save_flag:
        print(f"\nSuccessfully saved {num_imgs} images!")


def save_multipage_image(imgs_in, path_in, bigtiff=0,\
    compression=False, quiet_in=False):
    """ 
    Saves a list (or Numpy array) of images to the hard drive.
    The image stack will be saved specifically as a multipage
    TIFF file. To save the image stack as multiple, individual
    images, see save_image_seq(...).

    ---- INPUT ARGUMENTS ----
    [imgs_in]: A 3D Numpy array containing images, each represented by
        a Numpy array, which corresponds to OpenCV's single-channel 
        uint8 (grayscale) image data type. The shape of imgs_in should
        be (num_images, num_rows, num_cols).

    path_in: String that contains the file path of the image being 
        saved. This string should end with either '.tif' or '.tiff'. 
        Note, the file path will be normalized and made lowercase 
        (on Windows 10), so forward or backward slashes are acceptable.

    bigtiff: Integer that is either 0, 1, or 2. Changes the header and
        TIFF codec formatting. Select 0 for a standard TIFF file. If
        the file format will exceed 4 GB, then select either 1 or 2.
        Selecting 1 corresponds to using BigTIFF format, and selecting
        2 corresponds to using the native ImageJ Hyperstack format.

    compressions: Boolean that governs whether the image will be 
        compressed via ZLib. True to compress the multi-page TIFF
        image, and False to leave it as uncompressed.

    quiet_in: An optional boolean that is by default set to False. Set 
        to True to prevent outputting any messages.

    ---- RETURNED ---- 
    Nothing is returned    

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Image files are written
    to the hard drive. Strings may be printed to standard output.
    """

    # Make local copies
    imgs = imgs_in # No need for a deep copy. This image won't be altered

    # Forcing Python to create a new string variable in memory
    file_path = (path_in + '.')[:-1]
    file_path = os.path.normcase(file_path)

    quiet = quiet_in

    # Reduced list (actually tuple) of supported image file extensions
    file_ext3 = (".tif",)
    file_ext4 = (".tiff",)

    # Check for the correct file extension and save it
    if (file_path.lower()).endswith(file_ext3):
        file_ext = file_path[-4:] # Not actually used

    elif (file_path.lower()).endswith(file_ext4):
        file_ext = file_path[-5:] # Not actually used

    else:
        print(f"\nWarning: No supported file extension found in "\
            f"'{file_path}'\n    Defaulting to '.tif' file extension.")
        file_path = file_path + ".tif"

    num_imgs = imgs.shape[0]

    if not quiet:
        print(f"\nBeginning to save {num_imgs} images to:\n    {file_path}")

    # OpenCV's imwrite is supposed to automatically create a multi-page tiff
    # file when the filepath ends with ".tif" and a 3D Numpy array is given
    # as input. However, it does not work for me.
    #retval = cv.imwrite(file_path, img_list)

    # Sci-Kit Image does not officially support multi-page TIFF files, but
    # a number of other libraries do, such as PIL and TIFFFILE. Instead of
    # pulling in yet another dependency, I tried a workaround where I call
    # different plugin within Sci-Kit Image to perform the operation. As
    # it turns out, TIFFFILE automatically recognizes a 3D Numpy array and
    # will save it as a multi-page TIF file.


    # note: I think this is a somewhat cleaner way to set it up.
    # P.S. I wanted some python practice so I did this - certainly wasn't
    # necessary. There might still be a better way...
    if compression == True:
        compression = 'zlib'
    elif compression == False:
        compression = ''
    else:
        if not quiet:
            print("WARNING: 'compression' is unset. Defaulting to False")
        compression = ''
    
    if bigtiff == 0:
        bigtiff = False
        imagej = False
    elif bigtiff == 1:
        bigtiff = True
        imagej = False
    elif bigtiff == 2:
        bigtiff = False
        imagej = True
    else:
        if not quiet: 
            print(f"\nWARNING: bigtiff entry '{bigtiff}' is invalid!")
            print(f"\nDefaulting to bigtiff=0")
        bigtiff = False
        imagej = False

    bigtiff_compression_settings = {'bigtiff': bigtiff,\
                                    'imagej': imagej,\
                                    'compressionargs': compression}
   
    io.imsave(file_path, imgs, plugin='tifffile', \
              photometric='minisblack', **bigtiff_compression_settings)

    if not quiet:
        print(f"\nSuccessfully saved {num_imgs} images!")



import numpy as np
import cv2 as cv


def crop_img(img_in, roi_in, quiet_in=False):
    """
    Returns a cropped portion of the original image, which is positioned
    about the center of the original image. In other words, the anchor
    for the cropping operation is always located on the center pixel
    of the original image.

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    (roi_in): A tuple of two or four integers representing:
        Option a (centered crop, length 2): the number of rows and columns
          of pixels to be retained in the final image. The first
          integer corresponds to the number of rows, and the second to
          columns. Use negative numbers to prevent cropping in that 
          dimension. For example, (-1, 100) would keep all of the rows of
          the original image and crop the number of columns to a 100 
          pixels. Integers can be even or odd, and the cropping operation
          is always anchored about the center pixel of the original image.
        Option b (arbitrary crop, length 4): works just like array slicing
          where image is cropped rectangularly following bounds given as
          (row_min, row_max, column_min, column_max).

    quiet_in: A boolean that determiens if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing

    ---- RETURNED ---- 
    [img_crop]: The resultant image after cropping. img_crop is in the
        same format as img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings are 
    printed to standard output.
    """

    # ---- Start Local Copies ----
    img = img_in.copy() # Makes a proper (i.e., deep) copy
    roi = list(roi_in) # Region-of-interest (ROI)
    quiet = quiet_in
    # ---- End Start Local Copies ----

    # If not already, force the data to be of type integer
    for count, element in enumerate(roi):
        roi[count] = int(element)

    img_shape = img.shape # Tuple containing the number of rows and columns
    
    if len(roi) == 2: # Use a center anchor and number of rows and columns in roi
        
        if (roi[0] > img_shape[0]) or (roi[1] > img_shape[1]):
            if not quiet:
                print(f"\nWARNING: Cropped region is larger than the original "\
                    f"\nimage. Original Image Shape: {img_shape}"\
                    f"\nInput Crop Parameters:{roi}"\
                    f"\nUsing the original image size.")
            return img

        elif (roi[0] <= 0) and (roi[1] <= 0): # No cropping desired
            return img 

        # Keep the entire row or column if this is less than or equal to zero
        if roi[0] <= 0:
            roi[0] = img_shape[0]
        if roi[1] <= 0:
            roi[1] = img_shape[1]

        # Even integers in the ROI require special treatment
        r_even, c_even = False, False
        if roi[0] % 2 == 0:
            r_even = True
            roi[0] -= 1 # Make it odd for now, and correct for it later

        if roi[1] % 2 == 0:
            c_even = True
            roi[1] -= 1 # Make it odd for now, and correct for it later

        # Get the middle (row, col) indices of the original image. 
        if img_shape[0] % 2 == 0: # If even
            mid_r_index = int(img_shape[0]/2) - 1 # -1 b/c indices start at 0
        else: # Else odd
            mid_r_index = int(img_shape[0]/2) # int() rounds down, so OK

        if img_shape[1] % 2 == 0: # If even
            mid_c_index = int(img_shape[1]/2) - 1 # -1 b/c indices start at 0
        else: # Else odd
            mid_c_index = int(img_shape[1]/2) # int() rounds down, so OK

        mid_indices = (mid_r_index, mid_c_index) 

        # Number of pixels to keep above and below (or left and right of) the  
        # center. Should not need int(), but just in case.
        r_radius = int((roi[0] - 1)/2) # (Odd_Number - 1)/2 
        c_radius = int((roi[1] - 1)/2) # (Odd_Number - 1)/2 

        # List of lower/upper bounds on the indices to be kept
        r_bounds = [mid_indices[0] - r_radius, mid_indices[0] + r_radius + 1]
        c_bounds = [mid_indices[1] - c_radius, mid_indices[1] + c_radius + 1]

        # Middle coordinate was subtracted by one if even, so add extra index on
        # the upper end to ensure the correct length in the ROI
        if r_even: # If originally had even number of rows in ROI
            r_bounds[1] = r_bounds[1] + 1
        if c_even: # If originally had even number of columns in ROI
            c_bounds[1] = c_bounds[1] + 1
        crop_type = 'Centered'
        
    elif len(roi) == 4: # roi defines rmin,rmax,cmin,cmax
        # Check that the difference between min and max is less than or 
        # equal to the img size, and also that the max is smaller than or
        # equal to the image size.
        if ( (roi[1] - roi[0]) > img_shape[0]) or ( (roi[3] - roi[2]) > img_shape[1])\
            or (roi[1] > img_shape[0]) or (roi[3] > img_shape[1]):
            if not quiet:
                print(f"\nWARNING: Cropped region is larger than the original "\
                    f"\nimage. Original Image Shape: {img_shape}"\
                    f"\nInput Crop Parameters:{roi}"\
                    f"\nUsing the original image size.")
            return img

        elif (roi[0] <= 0) and (roi[1] <= 0) and (roi[2] <= 0) and (roi[3] <= 0): # No cropping desired
            return img 
    
        r_bounds = [roi[0], roi[1]]
        c_bounds = [roi[2], roi[3]]
        crop_type = 'Rectangle'
        
    else:
        print(f"\nWARNING: ROI parameters malformed (too few or too many)."
                    f"\nInput Crop Parameters:{roi}"\
                    f"\nUsing the original image size.")
        return img
        
    # Perform the actual slicing operation to crop the original image
    img_crop = img[r_bounds[0]:r_bounds[1], c_bounds[0]:c_bounds[1]]

    if not quiet:
        print("\nSuccessfully cropped the image:\n"\
            f"    Crop type: {crop_type}\n"\
            f"    Number of Rows: {img_crop.shape[0]} pixels\n"\
            f"    Number of Columns: {img_crop.shape[1]} pixels")

    # Return the cropped image
    return img_crop


def normalize_histogram(img_in, bounds_in=(0,255), quiet_in=False):
    """
    Perform a linear normalization of a grayscale digital image, also
    known as histogram stretching. The minimum pixel intensity will be 
    mapped to 0, and maximum pixel intensity to 255.

    ---- INPUT ARGUMENTS ----
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    (bounds_in): A tuple containing two integers that must range between
        0 and 255. The first integer is the intensity lower-bound, and
        the second integer is the intensity upper-bound. These intensity
        bounds will be used to truncate the original image's histogram.
        After normalization, any pixel that was originally below the 
        lower-bound will be black (i.e., 0). Conversely, any pixel that
        was originally above the upper-bound will be white (i.e., 255).

    quiet_in: A boolean that determiens if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing

    ---- RETURNED ----
    [img_new]: Same data structure as the input, img_in. However, the 
        returned image numpy array is stretched so that the intensity
        values span from 0 to 255.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings are 
    printed to standard output.
    """

    # ---- Start Local Copies ----
    img = img_in.copy() # Makes a proper (i.e., deep) copy
    low_bound = bounds_in[0]
    up_bound = bounds_in[1]
    quiet = quiet_in
    # ---- End Start Local Copies ----

    if low_bound == up_bound:
        low_bound = 0
        up_bound = 255

    elif low_bound > up_bound:
        temp_bound = low_bound
        low_bound = up_bound
        up_bound = temp_bound

    if up_bound > 255:
        up_bound = 255

    if low_bound < 0:
        low_bound = 0

    # Effectively threshold the upper and lower bounds of the image's
    # histogram based on the provided inputs
    img[img < low_bound] = low_bound
    img[img > up_bound] = up_bound

    # This is just a new view into img, does not make an actual memory copy
    img_new = img.copy()

    # OpenCV's normalize function will do this automatically
    cv.normalize(img, img_new, 0, 255, cv.NORM_MINMAX, cv.CV_8U)

    mean1_intensity = cv.mean(img_in)
    mean2_intensity = cv.mean(img_new)

    if not quiet:
        print("\nSuccessfully normalized the intensity histogram from"\
            " 0 to 255:\n"\
            f"    Mean Intensity of Original Image: {mean1_intensity[0]}\n"\
            f"    Mean Intensity of Normalized Image: {mean2_intensity[0]}")

    return img_new


def multi_morph(img_in, morph_params_in):
    """
    Applies a morphological operation (i.e., dilation and erosion
    transformations). Depending on the input parameters, this function
    can either perform just dilation transformation, or just erosion
    transformations. Alternatively, the user can perform either an open
    transformation or a close transformation. Note, an open
    transformation is some number of erosion-steps followed by some
    number of dilation-steps, and it can be helpful in removing noise.
    Conversely, a close transformation is  some number of dilation-steps
    followed by some number of erosion-steps, and it can be useful in
    closing small holes.

    ---- INPUT ARGUMENTS ----
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    [morph_params_in]: A list of parameters necessary to define the
        morphological operation to be applied. A more detailed
        explanation of each parameter in morph_params_in is given
        below,
            
            [flag_open_close, flag_rect_ellps, k_size, num_erode,
            num_dilate]

                flag_open_close: Either 0 or 1, and determines if the
                    erosion operations should occur prior to dilation
                    operations, or the other way around. If 0, then an 
                    open operation is done, which is erosion followed by 
                    dilation (removes noise). If 1, then a close 
                    operation is done, which is a dilation followed by
                    erosion (closes small holes). 

                flag_rect_ellps: Either 0 or 1. If 0, then a rectangular
                    kernel is used for the morphological operations. If
                    1, then an elliptical kernel is used.

                k_size: Kernel size for the morphological operations. If
                    less than 1, then the original image is returned.

                num_erode: Number of erosion operations to perform

                num_dilate: Number of dilation operations to perform

    ---- RETURNED ----
    [img_morph]: The resultant image after various erosion and dilation
        transformations is returned. The returned image has the same 
        data structure and data type as img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function.
    """

    # ---- Start Local Copies ----
    img = img_in.copy()
    morph_params = morph_params_in
    # ---- End Local Copies ----

    # 0 is open => erosion followed by dilation (removes noise)
    # 1 is close => dilation followed by erosion (closes small holes)
    flag_open_close = morph_params[0]

    # 0 is rectangular kernel, 1 is elliptical kernel
    flag_rect_ellps = morph_params[1]

    # Width (or diameter) of the kernel in pixels
    k_size = morph_params[2]

    # Number of iterations of the erode and dilation operations
    num_erode = morph_params[3]
    num_dilate = morph_params[4]

    # Show original image if kernel size is less than one, or the number
    # of iterations for both erosion and dilation are zero
    if k_size < 1:
        return img
    elif (num_dilate < 1) and (num_erode < 1):
        return img

    # Create the kernel
    if flag_rect_ellps: # Elliptical
        cur_kernel = cv.getStructuringElement(cv.MORPH_ELLIPSE, 
            (k_size,k_size))
    else: # Rectangular
        cur_kernel = cv.getStructuringElement(cv.MORPH_RECT, 
            (k_size,k_size))

    # Do the successive morphological operations
    if flag_open_close: # Close
        if (num_dilate >= 1) and (num_erode >= 1):
            img_morph = cv.dilate(img, cur_kernel, 
                iterations=num_dilate)
            img_morph = cv.erode(img_morph, cur_kernel, 
                iterations=num_erode)

        elif num_dilate < 1: # Only do a erosion
            img_morph = cv.erode(img, cur_kernel, 
                iterations=num_erode)

        else: # Only do an dilation
            img_morph = cv.dilate(img, cur_kernel, 
                iterations=num_dilate)

    else: # Open
        if (num_dilate >= 1) and (num_erode >= 1):
            img_morph = cv.erode(img, cur_kernel, 
                iterations=num_erode)
            img_morph = cv.dilate(img_morph, cur_kernel, 
                iterations=num_dilate)

        elif num_dilate < 1: # Only do a erosion
            img_morph = cv.erode(img, cur_kernel, 
                iterations=num_erode)

        else: # Only do an dilation
            img_morph = cv.dilate(img, cur_kernel, 
                iterations=num_dilate)

    return img_morph


def unsharp_mask(img_in, unsharp_params_in):
    """
    Sharpens an image using unsharp masking. This is achieved by
    subtracting a blurred image from the original, and then adding the
    result to the original image by some specified amount (see below), 
        sharpened = original + (original − blurred) X amount

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    [unsharp_params_in]: A list of parameters needed to perform the
        edge enhancement effect. Details of each parameter in 
        unsharp_params_in is given below,

            [cur_amount, k_size, fltr_type, std_dev]

                cur_amount: A percentage value of how much to blend
                    the sharpening effect onto the original value. It
                    corresponds to (amount)*100 in the equation above.
                    If zero, then no filter is applied.

                k_size: Kernel size for the blur filter. If zero, 
                    then no filter was applied. Must be odd.

                fltr_type: Either 0 or 1. If 0, then the blurred image
                    used in the procedure utilizes the Gaussian filter.
                    If 1, then the median filter is used.

                std_dev: Corresponds to the standard deviation used in
                    defining the Gaussian kernel for the Gaussian
                    filter. This parameter is ignored if the median
                    filter is used. If std_dev < 0, then the standard
                    deviation is automatically calculated.

    ---- RETURNED ---- 
    [img_sharp]: Returns the resultant image after performing the
        image processing procedures. img_sharp is in the same format as
        img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function.
    """

    # ---- Start Local Copies ----
    img = img_in.copy()
    unsharp_params = unsharp_params_in
    # ---- End Local Copies ----

    # As a percentage, this is a weight for adding images together. A 
    # typical value would be between 50% and 150%. Larger amount increases
    # the sharpening effect
    amount = unsharp_params[0]/100.0 # Convert out of a percentage

    # Related to kernel size of the blur filter. This affects the edge
    # thickness of the sharpened image
    radius = unsharp_params[1] # In pixels
    if radius % 2 == 0: # Must be odd for the blur filters
        radius += -1

    # Affects which blur filter is used. 0 does a Gaussian blur, while 1
    # does a median blur, which is more edge-preserving.
    fltr_type = unsharp_params[2]

    # Standard deviation for Guassian blur. This is ignored if using the
    # median blur filter. If less than 0, then this is automatically 
    # calculated
    gaus_sdev = unsharp_params[3]

    # Don't apply a sharp filter in these cases
    if amount <= 0:
        return img
    elif radius < 1:
        return img

    # Calculate the parameters for the Gaussian kernel
    k_size = (radius, radius) # Radius (i.e., kernel size)

    if gaus_sdev < 0:
        gaus_sdev = 0.3*((radius- 1)*0.5 - 1) + 0.8 # Standard deviation

    # Perform a blur filter
    if fltr_type:
        img_blur = cv.medianBlur(img, radius)
    else:
        img_blur = cv.GaussianBlur(img, k_size, gaus_sdev)

    img_sharp = cv.addWeighted(img, 1.0+amount, img_blur, -amount, 0)

    return img_sharp


def laplacian_sharp(img_in, sharp_params_in):
    """
    Sharpens an image using the edges detected by Laplacian.  The
    sharpening effect is achieved by subtracting the Laplacian image
    from the original (see below), 
        sharpened = original - (laplacian) X amount

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    [sharp_params_in]: A list of parameters needed to perform the
        edge enhancement effect. Details of each parameter in 
        sharp_params_in is given below,

            [cur_amount, blur_k_size, lap_k_size, fltr_type, 
            blur_std_dev]

                cur_amount: A percentage value of how much to blend
                    the sharpening effect onto the original value. It
                    corresponds to (amount)*100 in the equation above.
                    If zero, then no filter is applied.

                blur_k_size: Kernel size for the blur filter that is
                    applied prior to using the Laplacian kernel. If 
                    zero, then no blur filter is applied. Must be odd.

                lap_k_size: Kernel size for the Laplacian kernel used
                    to find the edges of the original image.

                fltr_type: Either 0 or 1. If 0, then the blurred image
                    used in the procedure utilizes the Gaussian filter.
                    If 1, then the median filter is used.

                blur_std_dev: Corresponds to the standard deviation used
                    in defining the Gaussian kernel for the Gaussian
                    blur. This parameter is ignored if the median blur
                    is used. If std_dev < 0, then the standard deviation
                    is automatically calculated.

    ---- RETURNED ---- 
    [img_sharp]: Returns the resultant image after performing the
        image processing procedures. img_sharp is in the same format as
        img_in.
        
    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function.
    """
    
    # ---- Start Local Copies ----
    img = img_in.copy()
    sharp_params = sharp_params_in
    # ---- End Local Copies ----

    # As a percentage, this is a weight for adding images together. A 
    # typical value would be between 50% and 150%. Larger amount increases
    # the sharpening effect
    amount = sharp_params[0]/100.0 # Convert out of a percentage

    # Related to kernel size of the Gaussian kernel. This affects the edge
    # thickness of the sharpened image
    blur_radius = sharp_params[1] # In pixels
    if blur_radius % 2 == 0: # Must be odd for the Gaussian blur
        blur_radius += -1

    # Radius of the kernel used to calculate the derivatives (Laplacian)
    lap_radius = sharp_params[2]
    if lap_radius % 2 == 0: 
        lap_radius += -1

    # Affects which blur filter is used. 0 does a Gaussian blur, while 1
    # does a median blur, which is more edge-preserving.
    fltr_type = sharp_params[3]

    # Standard deviation for Guassian blur. This is ignored if using the
    # median blur filter. If less than 0, this is automatically calculated
    gaus_sdev = sharp_params[4]

    # Don't apply a sharp filter in these cases
    if amount <= 0:
        return img
    elif (lap_radius < 1) or (blur_radius < 1):
        return img

    # Calculate the parameters for the Gaussian kernel
    blur_k_size = (blur_radius, blur_radius) # Radius (i.e., kernel size)

    if gaus_sdev < 0:
        gaus_sdev = 0.3*((blur_radius- 1)*0.5 - 1) + 0.8 # Standard deviation

    # Perform a blur filter
    if fltr_type:
        img_blur = cv.medianBlur(img, blur_radius)
    else:
        img_blur = cv.GaussianBlur(img, blur_k_size, gaus_sdev)

    # Perform the Laplacian filter to find the edges based on local gradients
    img_lap_64 = cv.Laplacian(img_blur, cv.CV_64F, ksize=lap_radius)

    img_sharp = cv.addWeighted(img, 1.0, img_lap_64, -amount, 0, 
        dtype=cv.CV_8U)

    return img_sharp


def canny_sharp(img_in, sharp_params_in):
    """
    Sharpens an image based on the Canny Edge Detection algorithm. The
    sharpening effect is achieved by subtracting the Canny-edges image
    from the original (see below), 
        sharpened = original - (canny) X amount
    Note, the Canny Edge Detection algorithm finds edges and marks them
    with discrete lines, so the resultant sharpened image will most
    likely not exhibit smooth gradients in contrast near the edges. 

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    [sharp_params_in]: A list of parameters needed to perform the
        edge enhancement effect. Details of each parameter in 
        sharp_params_in is given below,

            [k_size, cur_amount, thresh1, thresh2]

                k_size: Kernel size for the Canny edge algorithm. If
                    zero, then no filter was applied. Must be odd.

                cur_amount: A percentage value of how much to blend
                    the sharpening effect onto the original value. It
                    corresponds to (amount)*100 in the equation above.
                    If zero, then no filter is applied.

                thresh1: Related to the hystersis thresholding algorithm
                    in the Canny Edge Detection procedure. Related to 
                    minimum threshold used to determine what an edge is.

                thresh2: Related to the hystersis thresholding algorithm
                    in the Canny Edge Detection procedure. Related to 
                    maximum threshold used to determine what an edge is.

    ---- RETURNED ---- 
    [img_sharp]: Returns the resultant image after performing the
        image processing procedures. img_sharp is in the same format as
        img_in.
        
    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function.
    """
    
    # ---- Start Local Copies ----
    img = img_in.copy()
    sharp_params = sharp_params_in
    # ---- End Local Copies ----

    k_size = sharp_params[0]
    if k_size % 2 == 0: 
        k_size += -1

    amount = sharp_params[1]/100.0

    threshold1 = sharp_params[2]

    threshold2 = sharp_params[3]

    img_edges = cv.Canny(img, threshold1, threshold2, apertureSize=k_size, 
        L2gradient=True)

    img_sharp = cv.addWeighted(img, 1.0, img_edges, -amount, 0.0)

    return img_sharp


def global_equalize(img_in):
    """
    Enhances the contrast of a grayscale image by equalizing the 
    histogram. This is a global method, so the same parameters are
    utilized everywhere. For more information, vist "https://docs.
    opencv.org/4.2.0/d5/daf/tutorial_py_histogram_equalization.html"

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    ---- RETURNED ---- 
    [img_eq]: The equalized image is returned in the same data type and
        format at img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function.
    """

    # ---- Start Local Copies ----
    img = img_in.copy()
    # ---- End Local Copies ----

    img_eq = cv.equalizeHist(img)
    return img_eq


def adaptive_equalize(img_in, eq_params_in):
    """
    Enhances the contrast of a grayscale image by locally equalizing 
    the histogram. This is a locally adaptive method. For more 
    information, vist "https://docs.opencv.org/4.2.0/d5/daf/
    tutorial_py_histogram_equalization.html"

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    [eq_params_in]: A list of parameters necessary for the locally 
        adaptive equalization procedure. More details of each parameter
        in eq_params_in is given below,

            [blk_size, c_offset]
                
                blk_size: Size of a pixel neighborhood that is used to
                    calculate a threshold value for the current pixel.
                    Must be odd, and if zero, no thresholding is done.
                
                c_offset: A constant offset value applied to the mean
                    or weighted mean of the intensities


    ---- RETURNED ---- 
    [img_eq]: The equalized image is returned in the same data type and
        format at img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function.
    """

    # ---- Start Local Copies ----
    img = img_in.copy()
    eq_params = eq_params_in
    # ---- End Local Copies ----

    # Threshold for contrast limiting
    clip_limit = eq_params[0]

    # Size of grid for histogram equalization. Input image will be divided
    # into equally sized rectangular tiles. tileGridSize defines the number of
    # tiles in row and column.
    grid_size = eq_params[1]

    if grid_size < 1:
        return img

    tile_size = (grid_size, grid_size) 

    eq_obj = cv.createCLAHE(clip_limit, tile_size)
    img_eq = eq_obj.apply(img)

    return img_eq


def invert_binary_image(img_in, quiet_in=False):
    """
    Inverts a binarized (otherwise known as a black & white) image.
    Pixels that were black with intensity 0 become white with intensity
    255, and vice-versa.

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a black and white image. It is
        assumed  that the image is already of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    ---- RETURNED ---- 
    [img]: The inverted image is returned. It is in the same format and
        data type as img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function.
    """

    # ---- Start Local Copies ----
    img = img_in.copy()
    quiet = quiet_in
    # ---- End Local Copies ----

    img = cv.bitwise_not(img)

    if not quiet:
        print(f"\nSuccessfully inverted the image intensities.")

    return img


def calc_rel_density(img_in, invert_in=False, quiet_in=False):
    """
    Calculates the relative density of a binarized image, also known
    as a black and white image. By default, this function determines 
    the ratio of non-zero ratio pixels to the total number of pixels.
    Since white pixels are of intensity 255, another way to say this is
    that this function calculates portion of white pixels to total 
    pixels. A result of 1.0 implies the entire image is composed of
    non-zero pixels (i.e., white), and conversely, a result of 0.0 
    implies the entire image is black. However, it should be noted that
    the ratio of black pixels to total pixels can also be calculated,
    if desired.

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a black and white image. It is
        assumed  that the image is already of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    invert_in: A boolean that determines whether white pixels or black
        pixels are of interest. If set to False, the returned result
        is, (number of non-zero pixels)/(total number of pixels). If set
        to True, (number of zero pixels)/(total number of pixels).

    quiet_in: A boolean that determines if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing.

    ---- RETURNED ---- 
    rel_density: A float between 0.0 and 1.0 representing the relative
        density. Multiply by 100 to get a percentage.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Comments printed
    to standard output may occur.
    """
    quiet = quiet_in

    # Calculate number of non-black pixels
    num_white = cv.countNonZero(img_in)
    num_pixels = float(img_in.size) # Number of pixels in image

    if not invert_in:
        rel_density = num_white/num_pixels

    else:
        rel_density = (num_pixels - num_white)/num_pixels

    if not quiet:
        print(f"\nRelative Density of Final Image (from 0 to 1): "\
            f"{rel_density}")

    return rel_density


def create_blank_image(height_in, width_in, num_channels_in, 
    init_value_in=(0, 0, 0)):
    """
    Initializes a blank image with either 1 channel (grayscale) or 3
    channels (BGR format). The number of channels is also commonly
    referred to as the image depth. Images will be created using only
    the 'uint8' data type.

    ---- INPUT ARGUMENTS ----
    height_in: An integer representing the number of pixels in the 
        vertical direction (i.e., number of rows)

    width_in: An integer representing the number of pixels in the 
        horizontal direction (i.e., number of columns)

    num_channels_in: An integer that is either 1 or 3. Use 1 to create
        a single-channel image, which will be grayscale. Use 3 to 
        create a multi-channel image, which will be color. Recall that
        OpenCV uses Blue-Green-Red (BGR) format, rather than the 
        popular RGB format.

    (init_value_in): A tuple containing three integers. The blank image
        will be initialized to the color denoted by init_value_in. 
        Values should range between 0 and 255. If the desired image
        contains one channel, then the first value in init_value_in
        will be used to initialize the grayscale value.

    ---- RETURNED ---- 
    [img_blank]: The initialized image will be returned. The image will
        be based on OpenCV's NumPy data structure for images. The image
        will be of type 'uint8'.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function.
    """

    # Create local copies
    num_rows = height_in
    num_cols = width_in
    img_depth = num_channels_in
    img_color = init_value_in

    # Should be either depth of 1 (grayscale) or 3 (BGR)
    if (img_depth != 1) and (img_depth != 3):

        if img_depth < 1:
            img_depth = 1

        elif img_depth > 3:
            img_depth = 3

        else:
            img_depth = 1

    # If depth of 1, just need one intensity value. Using the first entry
    if img_depth == 1:
        img_color1 = img_color[0] 
        img_blank = (np.zeros((num_rows, num_cols), np.uint8))
        img_blank[:,:] = img_color1

    else: # Initialize the color for the BGR image
        if len(img_color) != 3:
            img_color = (0, 0, 0)

        img_blank = (np.zeros((num_rows, num_cols, 3), np.uint8))
        img_blank[:,:] = img_color

    return img_blank




    



import numpy as np
from scipy import spatial as spt

from . import volume_image_processing as vol

class BBox:
    """
    This is a simple class that stores a number of attributes that helps
    to define a 3D box, which is used later on to represent a bounding
    box for a 3D data set. The box is presumed to be oriented in some
    direction in 3D space, but a 3D rotation can be performed to align
    edges of the box with a new, local XYZ-coordinate system. The 
    rotation matrix, R_zyx, can be applied to the corners of the box
    to rotate them into this local coordinate system via,

        np.dot(R_zyx, np.transpose(corners))

    Note, this rotation matrix is constructed through intrinsic Euler
    angles defined by x_rad, y_rad, and z_rad. The length, width, and
    thickness of the box is given by x_len, y_len, and z_len. However,
    these lengths are NOT sorted since they are based on the local
    coordinate system and not their individual magnitudes.
    """

    def __init__(self, x_rad=0., y_rad=0., z_rad=0.):
        self.x_rad = x_rad # Angle about X-axis (radians)
        self.y_rad = y_rad # Angle about Y-axis (radians)
        self.z_rad = z_rad # Angle about Z-axis (radians)
        self.x_len = 0.0 # Length along local X-direction
        self.y_len = 0.0 # Length along local Y-direction
        self.z_len = 0.0 # Length along local Z-direction
        self.volume = 0.0
        self.center = np.array([0.0, 0.0, 0.0])
        self.corners = np.zeros((8,3)) # Coordinates of all 8 corners
        self.R_zyx = np.zeros((3,3)) # Rotation matrix 

    def calc_Rzyx(self):
        # Build a rotation matrix based on z-y'-x'' intrinsic
        # Euler angles.

        cx = np.cos(self.x_rad)
        sx = np.sin(self.x_rad)
        R_x = np.array([ 
            [1., 0., 0.],
            [0., cx, -sx],
            [0., sx, cx] ])

        cy = np.cos(self.y_rad)
        sy = np.sin(self.y_rad)
        R_y = np.array([
            [cy, 0., sy],
            [0., 1., 0.],
            [-sy, 0., cy] ])

        cz = np.cos(self.z_rad)
        sz = np.sin(self.z_rad)
        R_z = np.array([
            [cz, -sz, 0.],
            [sz, cz, 0.],
            [0., 0., 1.] ])

        self.R_zyx = np.matmul(R_z, np.matmul(R_y, R_x))
        return self.R_zyx

    def calc_volume(self):
        self.volume = self.x_len*self.y_len*self.z_len
        return self.volume

    def calc_center(self):
        self.center = np.mean(self.corners, axis=0)
        return self.center

    def calc_x_vec(self):
        b_x_vec = self.corners[0,:] - self.corners[3,:]
        b_x_vec = b_x_vec / np.sqrt(np.power(b_x_vec,2).sum())
        return b_x_vec

    def calc_y_vec(self):
        b_y_vec = self.corners[2,:] - self.corners[3,:]
        b_y_vec = b_y_vec / np.sqrt(np.power(b_y_vec,2).sum())
        return b_y_vec

    def calc_z_vec(self):
        b_z_vec = self.corners[7,:] - self.corners[3,:]
        b_z_vec = b_z_vec / np.sqrt(np.power(b_z_vec,2).sum())
        return b_z_vec

    def copy_to_self(self, other):
        self.x_rad = other.x_rad
        self.y_rad = other.y_rad
        self.z_rad = other.z_rad
        self.x_len = other.x_len 
        self.y_len = other.y_len
        self.z_len = other.z_len 
        self.volume = other.volume
        self.center = other.center.copy()
        self.corners = other.corners.copy()
        self.R_zyx = other.R_zyx.copy() 


class BBox2D:
    """
    Same as BBox above, but a 2D simplification.
    """

    def __init__(self, z_rad=0.):
        self.z_rad = z_rad # Angle about Z-axis (radians)
        self.x_len = 0.0 # Length along local X-direction
        self.y_len = 0.0 # Length along local Y-direction
        self.area = 0.0
        self.center = np.array([0.0, 0.0])
        self.corners = np.zeros((4,2)) # Coordinates of all 4 corners
        self.R_z = np.zeros((2,2)) # Rotation matrix 

    def calc_Rz(self):
        # Build a 2D rotation matrix
        cz = np.cos(self.z_rad)
        sz = np.sin(self.z_rad)
        R_z = np.array([
            [cz, -sz],
            [sz, cz] ])

        self.R_z = R_z
        return self.R_z

    def calc_area(self):
        self.area = self.x_len*self.y_len
        return self.area

    def calc_center(self):
        self.center = np.mean(self.corners, axis=0)
        return self.center

    def calc_x_vec(self):
        b_x_vec = self.corners[0,:] - self.corners[3,:]
        b_x_vec = b_x_vec / np.sqrt(np.power(b_x_vec,2).sum())
        return b_x_vec

    def calc_y_vec(self):
        b_y_vec = self.corners[2,:] - self.corners[3,:]
        b_y_vec = b_y_vec / np.sqrt(np.power(b_y_vec,2).sum())
        return b_y_vec

    def copy_to_self(self, other):
        self.z_rad = other.z_rad
        self.x_len = other.x_len 
        self.y_len = other.y_len
        self.area = other.area
        self.center = other.center.copy()
        self.corners = other.corners.copy()
        self.R_z = other.R_z.copy() 


def min_bounding_box(pnts_in, calc_hull=True, search=0):
    """
    This is a driver function to calculate a bounding box for the 3D
    data points defined by pnts_in. The bounding box can be arbitrarily
    oriented in 3D space; the objective is to find a bounding box that
    contains all of pnts_in that also results in the minimum volume of
    said bounding box. Depending on the level of accuracy and 
    computational efficiency, different search algorithms can be used.

    ---- INPUT ARGUMENTS ----
    [[pnts_in]]: A 2D numpy array containing all of the 3D points. For n
        points, then the shape should be [n,3]. The convex hull of this 
        data set can be calculated automatically.

    calc_hull: A boolean that determines whether the convex hull should
        be calculated and used for fitting the bounding box. Setting 
        this to True often speeds up performance for larger data sets.

    search: An integer that describes what type of bounding box search
        algorithm to use. If search equals:

        0: Use an exhaustive search. Although not necessarily the
            fastest, this method should find the best bounding box 
            within 1 degree of any arbitrary rotation. This method is 
            the default search algorithm.

        1: Perform a singular-value decomposition calculation on the 
            data set and use the eigenvector corresponding to the 
            minimum variation (i.e., smallest eigenvalue) as one of the
            directions of the bounding box. The other two directions are
            found by projecting the points onto a plane normal to this
            minimum principal direction, and then solving the 2D 
            bounding box problem using an exhaustive search. In 
            practice, this method usually is slightly more accurate
            than using the maximum variation (see search=2 below) while
            being just as computationally efficient. Performing the
            convex hull usually improves the accuracy for this search.

        2: Perform a singular-value decomposition calculation on the 
            data set and use the eigenvector corresponding to the 
            maximum variation (i.e., largest eigenvalue) as one of the
            directions of the bounding box. The other two directions are
            found by projecting the points onto a plane normal to this
            maximum principal direction, and then solving the 2D 
            bounding box problem using an exhaustive search. Performing
            the convex hull usually improves the accuracy for this
            search algorithm.

        3: Perform a singular-value decomposition calculation on the 
            data set and use all three eigenvectors as directions of
            the bounding box. Although this is the fastest method for
            large data sets, this is also the least reliable method in
            terms of accuracy. Performing the convex hull usually 
            improves the accuracy for this search algorithm.

        4: Find the longest distance between two points and define this
            direction to be major z-axis of the bounding box. The other
            two directions are found by projecting the points onto a 
            plane normal to this maximum principal direction, and then
            solving the 2D bounding box problem using an exhaustive 
            search.

    ---- RETURNED ---- 
    class BBox: See above for the definition of this class, which just
        contains a number of useful attributes.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the
    hard drive. Nothing is printed to the standard output stream.
    """

    num_pnts = pnts_in.shape[0]

    # Calculate the convex hull of the data set if desired
    if num_pnts > 4:
        if calc_hull:
            hull_obj = spt.ConvexHull(pnts_in)
            hull_pnts = pnts_in[hull_obj.vertices]
        else:
            hull_pnts = pnts_in
    else:
        hull_pnts = pnts_in


    if search is None:
        bbox_min = bbox_exhaustive_search(hull_pnts)

    elif search == 0:
        bbox_min = bbox_exhaustive_search(hull_pnts)

    elif search == 1:
        # Use PCA min
        bbox_min = bbox_svd_search(hull_pnts, algorithm="MIN")

    elif search == 2:
        # Use PCA max
        bbox_min = bbox_svd_search(hull_pnts, algorithm="MAX")

    elif search == 3:
        # Use all PCA directions
        bbox_min = bbox_svd_search(hull_pnts, algorithm="ALL")

    elif search == 4:
        # Set the longest-distance pair along one axis of the bbox
        bbox_min = bbox_LWT(hull_pnts)

    else:
        print("\nWARNING: Unrecognized search ID for calculating a bounding")
        print("box in function, min_bounding_box(...). Defaulting to an")
        print("exhaustive search for minimizing volume (search = 1).")
        bbox_min = bbox_exhaustive_search(hull_pnts)

    return bbox_min


def bbox_LWT(pnts_in):

    # Number of points
    n_pnts = pnts_in.shape[0] 
    dtype_in = pnts_in.dtype

    # Using the mean point to normalize the values.
    # Will need to translate the resultant bounding box later
    mean_pnt = np.round(np.mean(pnts_in, axis=0)).astype(dtype_in)
    pnts0 = pnts_in - mean_pnt

    # Candidate points for the longest distance between a pair of
    # points will lie on the convex hull.
    #pnts0 = pnts0[spt.ConvexHull(pnts0).vertices]

    # Get distances between each pair of candidate points
    dist_mat = spt.distance_matrix(pnts0, pnts0)

    # Get indices of pnts0 that are furthest apart
    L_pnt1_i, L_pnt2_i = np.unravel_index(dist_mat.argmax(), dist_mat.shape)

    L_pnt1 = pnts0[L_pnt1_i]
    L_pnt2 = pnts0[L_pnt2_i]
    L_vec = L_pnt2 - L_pnt1
    L_mag = np.sqrt(np.power(L_vec,2).sum())
    L_hat = L_vec/L_mag

    # Rotate the points so that the L vector is parallel to the
    # z-axis
    z_hat = np.array([0, 0, 1]) # [x, y, z]
    theta_local = np.arccos(np.clip(np.dot(L_hat, z_hat), -1.0, 1.0))

    if theta_local != 0.:
        rot_axis = np.cross(L_hat, z_hat)
        pnts_rot = vol.rodrigues_rot3(pnts0, rot_axis, theta_local,
            rot_center_in=np.array([0,0,0]))
    else:
        rot_axis = np.array([1, 0, 0])
        pnts_rot = pnts0.copy()

    R_zyx = vol.make_axis_angle_rot3_matrix(rot_axis, theta_local)
    
    # the bounding box is now aligned (or rather fixed) in the 
    # Z-direction. So, now we need to project the points onto 
    # the XY-plane and solve the 2D case.
    pnts_rot_2d = pnts_rot[:,0:2] # [n, 2] array now
    bbox_2d = bbox2D_exhaustive_search(pnts_rot_2d)

    z_rad = bbox_2d.z_rad
    cz = np.cos(z_rad)
    sz = np.sin(z_rad)
    R_z = np.array([
        [cz, -sz, 0.],
        [sz, cz, 0.],
        [0., 0., 1.] ])

    # Update pnts_rot
    pnts_rot =  np.dot(R_z, np.transpose(pnts_rot))
    pnts_rot = np.transpose(pnts_rot)

    # Effectively just need to add this additional Z-rotation
    # to the existing rotation matrix: Z-Y'-X''-Z'''
    R_zyx = np.matmul(R_z, R_zyx)

    bbox_3d = BBox()

    # Store the rotation matrix
    bbox_3d.R_zyx = R_zyx

    # Calculate the z-y'-x'' Euler (actually Tait-Bryan) angles
    # https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix
    R21 = R_zyx[1,0]
    R11 = R_zyx[0,0]
    R31 = R_zyx[2,0]
    R32 = R_zyx[2,1]
    R33 = R_zyx[2,2]

    # Due to round-off errors, this can be technically less than 0
    temp_radicand = 1 - R31*R31
    if temp_radicand < 0.0:
        temp_radicand = 0.0

    bbox_3d.z_rad = np.arctan2(R21, R11)
    bbox_3d.y_rad = np.arctan2(-R31, np.sqrt(temp_radicand))
    bbox_3d.x_rad = np.arctan2(R32, R33)

    # Get the axis-aligned bounding box 
    min_xyz = np.amin(pnts_rot, axis=0)
    max_xyz = np.amax(pnts_rot, axis=0)
    min_x = min_xyz[0]
    max_x = max_xyz[0]
    min_y = min_xyz[1]
    max_y = max_xyz[1]
    min_z = min_xyz[2]
    max_z = max_xyz[2]

    # Storing the corners temporarily in local coordinates
    bbox_3d.corners[0,:] = np.array([max_x, min_y, min_z])
    bbox_3d.corners[1,:] = np.array([max_x, max_y, min_z])
    bbox_3d.corners[2,:] = np.array([min_x, max_y, min_z])
    bbox_3d.corners[3,:] = np.array([min_x, min_y, min_z])
    bbox_3d.corners[4,:] = np.array([max_x, min_y, max_z])
    bbox_3d.corners[5,:] = np.array([max_x, max_y, max_z])
    bbox_3d.corners[6,:] = np.array([min_x, max_y, max_z])
    bbox_3d.corners[7,:] = np.array([min_x, min_y, max_z])

    bbox_3d.x_len = max_x - min_x
    bbox_3d.y_len = max_y - min_y
    bbox_3d.z_len = max_z - min_z
    bbox_vol = bbox_3d.calc_volume()

    # Update the corner coordinates to the global coordinate system
    bbox_corners = bbox_3d.corners
    R_xyz = np.transpose(R_zyx) # Transpose is the inverse rotation

    # Note on shapes: R_xyz[3,3] * (bbox_corners[8,3])^T 
    bbox_corners2 = np.dot(R_xyz, np.transpose(bbox_corners))
    bbox_corners2 = np.transpose(bbox_corners2)

    # Make sure to translate things back based on the mean point
    bbox_3d.corners = bbox_corners2.copy() + mean_pnt

    # Use the new corner coordinates to calculate the center point
    bbox_center = bbox_3d.calc_center()

    return bbox_3d


def bbox2D_LW(pnts_in):

    # Number of points
    n_pnts = pnts_in.shape[0]
    dtype_in = pnts_in.dtype

    # Using the mean point to normalize the values.
    # Will need to translate the resultant bounding box later
    mean_pnt = np.round(np.mean(pnts_in, axis=0)).astype(dtype_in)
    pnts0 = pnts_in - mean_pnt

    # Candidate points for the longest distance between a pair of
    # points will lie on the convex hull.
    #pnts0 = pnts0[spt.ConvexHull(pnts0).vertices]

    # Get distances between each pair of candidate points
    dist_mat = spt.distance_matrix(pnts0, pnts0)

    # Get indices of pnts0 that are furthest apart
    L_pnt1_i, L_pnt2_i = np.unravel_index(dist_mat.argmax(), dist_mat.shape)

    L_pnt1 = pnts0[L_pnt1_i]
    L_pnt2 = pnts0[L_pnt2_i]
    L_vec = L_pnt2 - L_pnt1
    L_mag = np.sqrt(np.power(L_vec,2).sum())
    L_hat = L_vec/L_mag

    # Rotate the points so that the L vector is parallel to the
    # x-axis
    x_hat = np.array([1, 0]) # [x, y]

    theta_local = np.arccos(np.clip(np.dot(L_hat, x_hat), -1.0, 1.0))
    rot_axis = np.cross(L_hat, x_hat) # Only z-component returned

    if rot_axis < 0:
         theta_local = -theta_local
    
    cz = np.cos(theta_local)
    sz = np.sin(theta_local)

    R_z = np.array([
        [cz, -sz],
        [sz, cz] ])

    R_z_inv = np.array([
        [cz, sz],
        [-sz, cz] ])

    if theta_local != 0:
        # pnts0 is shape [n,2], and R_z is shape [2,2]
        pnts_rot = np.matmul(R_z, np.transpose(pnts0))

        # Convert shape from [2,n] to [n,2]
        pnts_rot = np.transpose(pnts_rot)

    else:
        pnts_rot = pnts0.copy()

    # Now find the max and min row
    min_xy = np.amin(pnts_rot, axis=0)
    max_xy = np.amax(pnts_rot, axis=0)
    min_x = min_xy[0]
    max_x = max_xy[0]
    min_y = min_xy[1]
    max_y = max_xy[1]

    # Fill in a 2D bounding box object
    bbox_2d = BBox2D()
    bbox_2d.R_z = R_z
    bbox_2d.z_rad = theta_local

    # Storing the corners temporarily in local coordinates
    bbox_2d.corners[0,:] = np.array([max_x, min_y])
    bbox_2d.corners[1,:] = np.array([max_x, max_y])
    bbox_2d.corners[2,:] = np.array([min_x, max_y])
    bbox_2d.corners[3,:] = np.array([min_x, min_y])
    bbox_corners = bbox_2d.corners

    # Worth noting that, L = max_x - min_x, 
    # and, W = max_y - min_y
    bbox_2d.x_len = max_x - min_x
    bbox_2d.y_len = max_y - min_y
    bbox_area = bbox_2d.calc_area()

    # Note on shapes: R_z_inv[2,2] * (bbox_corners[4,2])^T 
    bbox_corners2 = np.dot(R_z_inv, np.transpose(bbox_corners))
    bbox_corners2 = np.transpose(bbox_corners2)

    # Make sure to translate things back based on the mean point
    bbox_2d.corners = bbox_corners2.copy() + mean_pnt

    # Use the new corner coordinates to calculate the center point
    bbox_center = bbox_2d.calc_center()

    return bbox_2d


def bbox_exhaustive_search(pnts_in):
    """
    Calculates an arbitrarily oriented bounding box for a set of 3D data
    points. This algorithm uses a brute-force search heuristic, so for
    very large data sets (greater than 10^5 points), this may be quite
    slow. The objective is to find an oriented 3D bounding box for
    which the volume of the bounding box is minimized. For large data
    sets, you may want to consider an algorithm inspired by principal
    component analysis that uses singular value decomposition, see for
    example bbox_svd_search(...). 

    ---- INPUT ARGUMENTS ----
    [[pnts_in]]: A 2D numpy array containing all of the 3D points. For n
        points, then the shape should be [n,3].

    ---- RETURNED ---- 
    class BBox: See above for the definition of this class, which just
        contains a number of useful attributes of a bounding box. 
        Notable attributes include the corners of the box, its  
        orientation (defined by a 3D rotation matrix), the volume, 
        the center point, its length, its width, and its depth.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the
    hard drive. Nothing is printed to the standard output stream.

    """
    pnts = pnts_in.copy() # [n,3] np-array

    # Using the mean point to normalize the values.
    # Will need to translate the resultant bounding box later
    mean_pnt = np.mean(pnts, axis=0)
    pnts = pnts - mean_pnt

    # Collect some platform information for 32-bit float data type
    fi_float = np.finfo(np.single)
    vol_min = fi_float.max

    # Initialize a the minimum bounding box object
    bbox_min = BBox()

    # Create arrays of angles (in degrees) that will be tested
    # Start with a coarse search, then later, refine the search
    ang_deg_step = 5.0
    ang_i = np.arange(0.0, 90.0, ang_deg_step) # Range for Z-rotations
    ang_j = np.arange(0.0, 90.0, ang_deg_step) # Range for Y-rotations
    ang_k = np.arange(0.0, 90.0, ang_deg_step) # Range for X-rotations

    for i_deg in ang_i: # Loop through Z-rotation angles
        for j_deg in ang_j: # Loop through Y-rotation angles
            for k_deg in ang_k: # Loop through X-rotation angles

                z_rad = np.deg2rad(i_deg)
                y_rad = np.deg2rad(j_deg)
                x_rad = np.deg2rad(k_deg)

                cur_bbox = BBox(x_rad, y_rad, z_rad)
                R_zyx = cur_bbox.calc_Rzyx()

                # Rotate the convex hull points
                # Note on shapes: R_zyx[3,3] * (pnts[n,3])^T 
                pnts_rot =  np.dot(R_zyx, np.transpose(pnts))

                # Convert shape from [3,n] to [n,3]
                pnts_rot = np.transpose(pnts_rot)

                # Get the axis-aligned bounding box 
                min_xyz = np.amin(pnts_rot, axis=0)
                max_xyz = np.amax(pnts_rot, axis=0)
                min_x = min_xyz[0]
                max_x = max_xyz[0]
                min_y = min_xyz[1]
                max_y = max_xyz[1]
                min_z = min_xyz[2]
                max_z = max_xyz[2]

                # Storing the corners temporarily in local coordinates
                cur_bbox.corners[0,:] = np.array([max_x, min_y, min_z])
                cur_bbox.corners[1,:] = np.array([max_x, max_y, min_z])
                cur_bbox.corners[2,:] = np.array([min_x, max_y, min_z])
                cur_bbox.corners[3,:] = np.array([min_x, min_y, min_z])
                cur_bbox.corners[4,:] = np.array([max_x, min_y, max_z])
                cur_bbox.corners[5,:] = np.array([max_x, max_y, max_z])
                cur_bbox.corners[6,:] = np.array([min_x, max_y, max_z])
                cur_bbox.corners[7,:] = np.array([min_x, min_y, max_z])

                cur_bbox.x_len = max_x - min_x
                cur_bbox.y_len = max_y - min_y
                cur_bbox.z_len = max_z - min_z
                cur_vol = cur_bbox.calc_volume()

                # Seeking out the bounding box with smallest volume
                if cur_vol < vol_min:
                    vol_min = cur_vol
                    bbox_min.copy_to_self(cur_bbox)
                    ijk_min = (i_deg, j_deg, k_deg)

    # Now refine the search near the previously found minimum BBox
    if (ijk_min[0]-5.0) < 0:
        i_deg_left = 0
    else:
        i_deg_left = ijk_min[0]-5.0

    if (ijk_min[0]+5.0) > 90:
        i_deg_right = 90
    else:
        i_deg_right = ijk_min[0]+5.0

    if (ijk_min[1]-5.0) < 0:
        j_deg_left = 0
    else:
        j_deg_left = ijk_min[1]-5.0

    if (ijk_min[1]+5.0) > 90:
        j_deg_right = 90
    else:
        j_deg_right = ijk_min[1]+5.0

    if (ijk_min[2]-5.0) < 0:
        k_deg_left = 0
    else:
        k_deg_left = ijk_min[2]-5.0

    if (ijk_min[2]+5.0) > 90:
        k_deg_right = 90
    else:
        k_deg_right = ijk_min[2]+5.0

    ang_deg_step = 1.0
    ang_i = np.arange(i_deg_left, i_deg_right, ang_deg_step) 
    ang_j = np.arange(j_deg_left, j_deg_right, ang_deg_step) 
    ang_k = np.arange(k_deg_left, k_deg_right, ang_deg_step) 

    for i_deg in ang_i: # Loop through Z-rotation angles
        for j_deg in ang_j: # Loop through Y-rotation angles
            for k_deg in ang_k: # Loop through X-rotation angles

                z_rad = np.deg2rad(i_deg)
                y_rad = np.deg2rad(j_deg)
                x_rad = np.deg2rad(k_deg)

                cur_bbox = BBox(x_rad, y_rad, z_rad)
                R_zyx = cur_bbox.calc_Rzyx()

                # Rotate the convex hull points
                # Note on shapes: R_zyx[3,3] * (pnts[n,3])^T 
                pnts_rot =  np.dot(R_zyx, np.transpose(pnts))

                # Convert shape from [3,n] to [n,3]
                pnts_rot = np.transpose(pnts_rot)

                # Get the axis-aligned bounding box 
                min_xyz = np.amin(pnts_rot, axis=0)
                max_xyz = np.amax(pnts_rot, axis=0)
                min_x = min_xyz[0]
                max_x = max_xyz[0]
                min_y = min_xyz[1]
                max_y = max_xyz[1]
                min_z = min_xyz[2]
                max_z = max_xyz[2]

                # Storing the corners temporarily in local coordinates
                cur_bbox.corners[0,:] = np.array([max_x, min_y, min_z])
                cur_bbox.corners[1,:] = np.array([max_x, max_y, min_z])
                cur_bbox.corners[2,:] = np.array([min_x, max_y, min_z])
                cur_bbox.corners[3,:] = np.array([min_x, min_y, min_z])
                cur_bbox.corners[4,:] = np.array([max_x, min_y, max_z])
                cur_bbox.corners[5,:] = np.array([max_x, max_y, max_z])
                cur_bbox.corners[6,:] = np.array([min_x, max_y, max_z])
                cur_bbox.corners[7,:] = np.array([min_x, min_y, max_z])

                cur_bbox.x_len = max_x - min_x
                cur_bbox.y_len = max_y - min_y
                cur_bbox.z_len = max_z - min_z
                cur_vol = cur_bbox.calc_volume()

                # Seeking out the bounding box with smallest volume
                if cur_vol < vol_min:
                    vol_min = cur_vol
                    bbox_min.copy_to_self(cur_bbox)
                    ijk_min = (i_deg, j_deg, k_deg)

    # With minimum bbox found, need to update the coordinates
    bbox_corners = bbox_min.corners
    R_zyx = bbox_min.R_zyx
    R_xyz = np.transpose(R_zyx) # Transpose is the inverse rotation

    # Note on shapes: R_xyz[3,3] * (bbox_corners[8,3])^T 
    bbox_corners2 = np.dot(R_xyz, np.transpose(bbox_corners))
    bbox_corners2 = np.transpose(bbox_corners2)

    # Make sure to translate things back based on the mean point
    bbox_min.corners = bbox_corners2.copy() + mean_pnt

    # Use the new corner coordinates to calculate the center point
    bbox_center = bbox_min.calc_center()

    return bbox_min


def bbox2D_exhaustive_search(pnts_in):
    """
    Calculates an arbitrarily oriented bounding box for a set of 2D data
    points. This algorithm uses a brute-force search heuristic, so for
    very large data sets, this may be rather slow. The objective is to
    find an oriented 2D bounding box for which the area of the bounding
    box is minimized.

    ---- INPUT ARGUMENTS ----
    [[pnts_in]]: A 2D numpy array containing all of the 2D points. For n
        points, then the shape should be [n,2].

    ---- RETURNED ---- 
    class BBox2D: See above for the definition of this class, which just
        contains a number of useful attributes of a 2D bounding box. 
        Notable attributes include the corners of the box, its  
        orientation (defined by a 2D rotation matrix), the area, 
        the center point, its length, and its width.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the
    hard drive. Nothing is printed to the standard output stream.
    """

    pnts = pnts_in.copy() # [n,2] np-array

    # Using the mean point to normalize the values.
    # Will need to translate the resultant bounding box later
    mean_pnt = np.mean(pnts, axis=0)
    pnts = pnts - mean_pnt

    # Collect some platform information for 32-bit float data type
    fi_float = np.finfo(np.single)
    area_min = fi_float.max

    # Initialize a the minimum bounding box object
    bbox_min = BBox2D()

    # Create arrays of angles (in degrees) that will be tested
    # Start with a coarse search, then later, refine the search
    ang_deg_step = 5.0
    ang_i = np.arange(0.0, 90.0, ang_deg_step) # Range for Z-rotations

    for i_deg in ang_i: # Loop through Z-rotation angles

        z_rad = np.deg2rad(i_deg)

        cur_bbox = BBox2D(z_rad)
        R_z = cur_bbox.calc_Rz()

        # Rotate the convex hull points
        # Note on shapes: R_z[2,2] * (pnts[n,2])^T 
        pnts_rot =  np.dot(R_z, np.transpose(pnts))

        # Convert shape from [2,n] to [n,2]
        pnts_rot = np.transpose(pnts_rot)

        # Get the axis-aligned bounding box 
        min_xy = np.amin(pnts_rot, axis=0)
        max_xy = np.amax(pnts_rot, axis=0)
        min_x = min_xy[0]
        max_x = max_xy[0]
        min_y = min_xy[1]
        max_y = max_xy[1]

        # Storing the corners temporarily in local coordinates
        cur_bbox.corners[0,:] = np.array([max_x, min_y])
        cur_bbox.corners[1,:] = np.array([max_x, max_y])
        cur_bbox.corners[2,:] = np.array([min_x, max_y])
        cur_bbox.corners[3,:] = np.array([min_x, min_y])

        cur_bbox.x_len = max_x - min_x
        cur_bbox.y_len = max_y - min_y
        cur_area = cur_bbox.calc_area()

        # Seeking out the bounding box with smallest volume
        if cur_area < area_min:
            area_min = cur_area
            bbox_min.copy_to_self(cur_bbox)
            i_min = i_deg

    # Now refine the search near the previously found minimum BBox
    if (i_min-5.0) < 0:
        i_deg_left = 0
    else:
        i_deg_left = i_min-5.0

    if (i_min+5.0) > 90:
        i_deg_right = 90
    else:
        i_deg_right = i_min+5.0

    ang_deg_step = 1.0
    ang_i = np.arange(i_deg_left, i_deg_right, ang_deg_step) 

    for i_deg in ang_i: # Loop through Z-rotation angles

        z_rad = np.deg2rad(i_deg)

        cur_bbox = BBox2D(z_rad)
        R_z = cur_bbox.calc_Rz()

        # Rotate the convex hull points
        # Note on shapes: R_z[2,2] * (pnts[n,2])^T 
        pnts_rot =  np.dot(R_z, np.transpose(pnts))

        # Convert shape from [2,n] to [n,2]
        pnts_rot = np.transpose(pnts_rot)

        # Get the axis-aligned bounding box 
        min_xy = np.amin(pnts_rot, axis=0)
        max_xy = np.amax(pnts_rot, axis=0)
        min_x = min_xy[0]
        max_x = max_xy[0]
        min_y = min_xy[1]
        max_y = max_xy[1]

        # Storing the corners temporarily in local coordinates
        cur_bbox.corners[0,:] = np.array([max_x, min_y])
        cur_bbox.corners[1,:] = np.array([max_x, max_y])
        cur_bbox.corners[2,:] = np.array([min_x, max_y])
        cur_bbox.corners[3,:] = np.array([min_x, min_y])

        cur_bbox.x_len = max_x - min_x
        cur_bbox.y_len = max_y - min_y
        cur_area = cur_bbox.calc_area()

        # Seeking out the bounding box with smallest volume
        if cur_area < area_min:
            area_min = cur_area
            bbox_min.copy_to_self(cur_bbox)
            i_min = i_deg

    # With minimum bbox found, need to update the coordinates
    bbox_corners = bbox_min.corners
    R_z = bbox_min.R_z
    R_z_inv = np.transpose(R_z) # Transpose is the inverse rotation

    # Note on shapes: R_z_inv[2,2] * (bbox_corners[4,2])^T 
    bbox_corners2 = np.dot(R_z_inv, np.transpose(bbox_corners))
    bbox_corners2 = np.transpose(bbox_corners2)

    # Make sure to translate things back based on the mean point
    bbox_min.corners = bbox_corners2.copy() + mean_pnt

    # Use the new corner coordinates to calculate the center point
    bbox_center = bbox_min.calc_center()

    return bbox_min


def bbox_svd_search(pnts_in, algorithm="MIN"):
    """
    Utilize singular value decomposition (SVD) to calculate the singular
    values of the 3D points given by pnts_in. Some or all of the 
    resultant eigenvectors are used to define the bounding box, 
    depending on the string used to define the algorithm parameter. In
    practice, algorithm = "MIN" will usually provide a more accurate
    bounding box compared to "MAX" in terms of finding an arbitrary 
    rotation that minimizes the volume of the bounding box. However, 
    these two methods are more computationally intensive than "ALL", 
    which is the fastest and least reliably accurate method. Note, a
    bounding box that is within 1.0 deg of rotational accuracy can be 
    found using the slower, exhaustive search given by the function
    above, bbox_exhaustive_search(...).

     ---- INPUT ARGUMENTS ----
    [[pnts_in]]: A 2D numpy array containing all of the 3D points. For n
        points, then the shape should be [n,3].

    algorithm: A string that is either "MIN", "MAX", or "ALL". The 
        default value is "MIN". When set to "MIN", the local z-axis of
        the bounding box will be aligned with direction from singular
        value decomposition (SVD) that corresponds to the minimum 
        variance. When set to "MAX", the local z-axis of the bounding
        box will be aligned with the direction from SVD that corresonds
        to the maximum variance. For either "MIN" or "MAX", only one
        direction from SVD is used to define the bounding box; the 
        remaining two directions are found from solving the 2D bounding
        box problem by projecting the data to the plane normal to the
        direction kept from SVD. When set to "ALL", all three directions
        from SVD are used to define the bounding box.

    ---- RETURNED ---- 
    class BBox: See above for the definition of this class, which just
        contains a number of useful attributes of a bounding box. 
        Notable attributes include the corners of the box, its  
        orientation (defined by a 3D rotation matrix), the volume, 
        the center point, its length, its width, and its depth.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the
    hard drive. Nothing is printed to the standard output stream.
    """
    
    pnts = pnts_in.copy()
    alg = algorithm.upper() # MIN, MAX, or ALL

    # Using the mean point to normalize the values.
    # Will need to translate the resultant bounding box later
    mean_pnt = np.mean(pnts, axis=0)
    pnts = pnts - mean_pnt

    # Use SVD to calculate the principal directions
    #  pnts: [n, 3]
    #  u: [n, n]
    #  s: [3]   <--- Singular values sorted in descending order
    #  vh: [3, 3]
    # The rows of vh are the eigenvectors of pnts^T * pnts
    # The columns of u are the eigenvectors of pnts * pnts^T
    u, s, vh = np.linalg.svd(pnts)

    # Note, the following code calculates pnts via the equation:
    #    pnts = u*s_mat*vh
    #
    # ---- Uncomment Below To Activate Code -----
    #
    #s_mat = np.zeros( (u.shape[0], vh.shape[0]) )
    #for m, cur_s in enumerate(s):
    #    s_mat[m,m] = cur_s
    #    
    #pnts_copy = np.matmul(u, np.matmul(s_mat, vh) )

    eig_vec1 = vh[0,:]
    eig_vec2 = vh[1,:]
    eig_vec3 = vh[2,:]

    mag1 = np.sqrt( np.power(eig_vec1[0], 2) + 
                    np.power(eig_vec1[1], 2) +
                    np.power(eig_vec1[2], 2) )
    eig_vec1 = eig_vec1/mag1

    mag2 = np.sqrt( np.power(eig_vec2[0], 2) + 
                    np.power(eig_vec2[1], 2) +
                    np.power(eig_vec2[2], 2) )
    eig_vec2 = eig_vec2/mag2

    mag3 = np.sqrt( np.power(eig_vec3[0], 2) + 
                    np.power(eig_vec3[1], 2) +
                    np.power(eig_vec3[2], 2) )
    eig_vec3 = eig_vec3/mag3

    if alg == "ALL":
        # Let us assume the bounding box will align its local x'-axis
        # to eig_vec1, y'-axis to eig_vec2, and z'-axis to eig_vec3.

        # The bbox x-vector expressed in global coordinates
        x_vec_g = eig_vec1 

        # The bbox y-vector expressed in global coordinates
        y_vec_g = eig_vec2

        # The bbox z-vector expressed in global coordinates
        z_vec_g = np.cross(x_vec_g, y_vec_g)  

    elif alg == "MAX":
        # In this case, the bounding box will align its local z'-axis
        # to eig_vec1. The remaining local axes can be arbitrarily 
        # chosen for the time being.

        # The bbox z-vector expressed in global coordinates
        z_vec_g = eig_vec1 

        # The bbox y-vector expressed in global coordinates
        y_vec_g = eig_vec2

        # The bbox x-vector expressed in global coordinates
        x_vec_g = np.cross(y_vec_g, z_vec_g)

    else: # "MIN"
        # In this case, the bounding box will align its local z'-axis
        # to eig_vec3. The remaining local axes can be arbitrarily 
        # chosen for the time being.

        # The bbox z-vector expressed in global coordinates
        z_vec_g = eig_vec3 

        # The bbox y-vector expressed in global coordinates
        y_vec_g = eig_vec2

        # The bbox x-vector expressed in global coordinates
        x_vec_g = np.cross(y_vec_g, z_vec_g)

    # Make the rotation matrix
    R_zyx = np.array([
        [x_vec_g[0], x_vec_g[1], x_vec_g[2] ],
        [y_vec_g[0], y_vec_g[1], y_vec_g[2] ],
        [z_vec_g[0], z_vec_g[1], z_vec_g[2] ] ])

    # Rotate the points and work in the local coordinate system
    # Note on shapes: R_zyx[3,3] * (pnts[n,3])^T 
    pnts_rot =  np.dot(R_zyx, np.transpose(pnts))

    # Convert shape from [3,n] to [n,3]
    pnts_rot = np.transpose(pnts_rot)

    if alg != "ALL":
        # For either the "MIN" or "MAX" case, the bounding box is now
        # aligned (or rather fixed) in the Z-direction. So, now we need
        # to project the points onto the XY-plane and solve the 2D case
        pnts_rot_2d = pnts_rot[:,0:2] # [n, 2] array now
        bbox_2d = bbox2D_exhaustive_search(pnts_rot_2d)

        z_rad = bbox_2d.z_rad
        cz = np.cos(z_rad)
        sz = np.sin(z_rad)
        R_z = np.array([
            [cz, -sz, 0.],
            [sz, cz, 0.],
            [0., 0., 1.] ])

        # Update pnts_rot
        pnts_rot =  np.dot(R_z, np.transpose(pnts_rot))
        pnts_rot = np.transpose(pnts_rot)

        # Effectively just need to add this additional Z-rotation
        # to the existing rotation matrix: Z-Y'-X''-Z'''
        R_zyx = np.matmul(R_z, R_zyx)

    # Initialize the BBox object that will be returned
    bbox_min = BBox()

    # Store the rotation matrix
    bbox_min.R_zyx = R_zyx

    # Calculate the z-y'-x'' Euler (actually Tait-Bryan) angles
    # https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix
    R21 = R_zyx[1,0]
    R11 = R_zyx[0,0]
    R31 = R_zyx[2,0]
    R32 = R_zyx[2,1]
    R33 = R_zyx[2,2]

    # Due to round-off errors, this can be technically less than 0
    temp_radicand = 1 - R31*R31
    if temp_radicand < 0.0:
        temp_radicand = 0.0

    bbox_min.z_rad = np.arctan2(R21, R11)
    bbox_min.y_rad = np.arctan2(-R31, np.sqrt(temp_radicand))
    bbox_min.x_rad = np.arctan2(R32, R33)

    # Get the axis-aligned bounding box 
    min_xyz = np.amin(pnts_rot, axis=0)
    max_xyz = np.amax(pnts_rot, axis=0)
    min_x = min_xyz[0]
    max_x = max_xyz[0]
    min_y = min_xyz[1]
    max_y = max_xyz[1]
    min_z = min_xyz[2]
    max_z = max_xyz[2]

    # Storing the corners temporarily in local coordinates
    bbox_min.corners[0,:] = np.array([max_x, min_y, min_z])
    bbox_min.corners[1,:] = np.array([max_x, max_y, min_z])
    bbox_min.corners[2,:] = np.array([min_x, max_y, min_z])
    bbox_min.corners[3,:] = np.array([min_x, min_y, min_z])
    bbox_min.corners[4,:] = np.array([max_x, min_y, max_z])
    bbox_min.corners[5,:] = np.array([max_x, max_y, max_z])
    bbox_min.corners[6,:] = np.array([min_x, max_y, max_z])
    bbox_min.corners[7,:] = np.array([min_x, min_y, max_z])

    bbox_min.x_len = max_x - min_x
    bbox_min.y_len = max_y - min_y
    bbox_min.z_len = max_z - min_z
    cur_vol = bbox_min.calc_volume()

    # Axis-aligned local bbox found; need to update the corner 
    # coordinates to the global coordinate system
    bbox_corners = bbox_min.corners
    R_xyz = np.transpose(R_zyx) # Transpose is the inverse rotation

    # Note on shapes: R_xyz[3,3] * (bbox_corners[8,3])^T 
    bbox_corners2 = np.dot(R_xyz, np.transpose(bbox_corners))
    bbox_corners2 = np.transpose(bbox_corners2)

    # Make sure to translate things back based on the mean point
    bbox_min.corners = bbox_corners2.copy() + mean_pnt

    # Use the new corner coordinates to calculate the center point
    bbox_center = bbox_min.calc_center()

    return bbox_min


# Building the package is done using:
#   python setup.py bdist_wheel sdist
#
# This requires pip, wheel, setuptools, numpy, and Cython, and the
# appropriate C compiler for you python system. For windows, see the
# Python documentation:
# https://wiki.python.org/moin/WindowsCompilers
#
# After successfully createing a .whl file, it can be installed
# locally using:
#    pip install /path/to/file.whl

import os
from setuptools import setup, find_packages, Extension

try:
    import numpy
except ImportError:  # We do not have numpy installed
    os.system("pip install numpy")

try:
    from Cython.Build import cythonize
except ImportError:  # We do not have Cython installed
    os.system("pip install Cython")

from Cython.Build import cythonize
import numpy

extensions = [
    Extension(
        "imppy3d.im3_processing",
        ["src/imppy3d/cython/im3_processing.pyx"],
        include_dirs=[numpy.get_include()]
    )
]

setup(
    name='imppy3d',
    version='1.0.2',
    author='Newell Moser',  
    author_email='newell.moser@nist.gov',  
    description='IMPPY3D: A library for processing 3D image stacks',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/usnistgov/imppy3d/',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    setup_requires=[
        "cython", 
        "numpy",
    ],
    install_requires=[
        'numpy == 1.26.*',
        'scipy == 1.11.*',
        'matplotlib == 3.8.*',
        'scikit-image == 0.20.*',
        'opencv-python == 4.6.*',
        'vtk == 9.2.*',
        'meshio == 5.3.*',
        'imageio >= 2.22',
        'imageio-ffmpeg >= 0.4',
        'pillow >= 9',
        'tifffile >= 2022',
        'imagecodecs >= 2022.7',
    ],
    ext_modules=cythonize(
        extensions,
        compiler_directives={'language_level' : "3"}
    ),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Cython",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires='>=3.9',
)